import gekim as gk
from gekim.fields.bio.enzyme.inhib import irrev as ii 
import numpy as np
import matplotlib.pyplot as plt
from scipy.integrate import solve_ivp

#TODO: add loop case
#TODO: add negative feedback case

# Test whether gekim NState systems yield the same output as hardcoded systems
# To add a new test case, add a new scheme to the schemes dictionary and make new hardcoded system class, then add it to compare. 
    # they need to share species with vani 3-state covlig systems, for simplicity

# User option to plot time course of NState systems (will still plot on mismatch)
PLOT=False

# Params
Kd = 10 #nM, koff/kon
koff = 0.0001 #~1/15 min
kon = koff/Kd
concI0,concE0=100,1
kinactf=0.001
kinactb=0
t = np.linspace(kon, 10000, 10000)

# Define gekim schemes
schemes={}
schemes["3S_vani"] = {
    "species": {
        "I": {"y0": concI0, "label": r"I"},
        "E": {"y0": concE0, "label": r"E"},
        "E_I": {"y0": 0, "label": r"E${\cdot}$I"},
        "EI": {"y0": 0, "label": r"E-I"},
    },
    "transitions": {
        "k1": {"k": kon, "source": ["E", "I"], "target": ["E_I"], "label": r"$k_{on}$"},
        "k2": {"k": koff, "source": ["E_I"], "target": ["E", "I"], "label": r"$k_{off}$"},
        "k3": {"k": kinactf, "source": ["E_I"], "target": ["EI"]}, #irrev step
        "k4": {"k": kinactb, "source": ["EI"], "target": ["E_I"]},
    },
}
schemes["3S_mod1.1"] = {
    "species": {
        "I": {"y0": concI0, "label": r"I"},
        "E": {"y0": concE0, "label": r"E"},
        "E_I": {"y0": 0, "label": r"E${\cdot}$I"},
        "EI": {"y0": 0, "label": r"E-I"},
    },
    "transitions": {
        "k1": {"k": kon, "source": ["E", "2.0I"], "target": ["E_I"], "label": r"$k_{on}$"},
        "k2": {"k": koff, "source": ["E_I"], "target": ["E", "2.0I"], "label": r"$k_{off}$"},
        "k3": {"k": kinactf, "source": ["3E_I"], "target": ["7EI"]}, #irrev step
        "k4": {"k": kinactb, "source": ["7EI"], "target": ["3E_I"]},
    },
}
schemes["3S_mod1.2"] = {
    "species": {
        "I": {"y0": concI0, "label": r"I"},
        "E": {"y0": concE0, "label": r"E"},
        "E_I": {"y0": 0, "label": r"E${\cdot}$I"},
        "EI": {"y0": 0, "label": r"E-I"},
    },
    "transitions": {
        "k1": {"k": kon, "source": ["E", "I", "I"], "target": ["E_I"], "label": r"$k_{on}$"},
        "k2": {"k": koff, "source": ["E_I"], "target": ["E", "I", "I"], "label": r"$k_{off}$"},
        "k3": {"k": kinactf, "source": ["E_I","E_I","E_I"], "target": ["7EI"]}, #irrev step
        "k4": {"k": kinactb, "source": ["7EI"], "target": ["E_I","E_I","E_I"]},
    },
}
schemes["3S_mod2.1"] = {
    "species": {
        "I": {"y0": concI0, "label": r"I"},
        "E": {"y0": concE0, "label": r"E"},
        "E_I": {"y0": 0, "label": r"E${\cdot}$I"},
        "EI": {"y0": 0, "label": r"E-I"},
    },
    "transitions": {
        "k1": {"k": kon, "source": ["2E", "I"], "target": ["E_I","E"], "label": r"$k_{on}$"},
        "k2": {"k": koff, "source": ["E_I","E"], "target": ["2E", "I"], "label": r"$k_{off}$"},
        "k3": {"k": kinactf, "source": ["E_I"], "target": ["EI"]}, #irrev step
        "k4": {"k": kinactb, "source": ["EI"], "target": ["E_I"]},
    },
}

schemes["3S_osc1"] = {
    "species": {
        "I": {"y0": concI0*1e-2, "label": r"I"},
        "E": {"y0": concE0, "label": r"E"},
        "E_I": {"y0": 0, "label": r"E${\cdot}$I"},
        "EI": {"y0": 0, "label": r"E-I"},
    },
    "transitions": {
        "k1": {"k": kon*8e1, "source": ["E", "0I"], "target": ["E_I"]},
        "k2": {"k": koff*5e2, "source": ["E_I"], "target": ["-E", "0I","E_I"]},
        "k3": {"k": kinactf*1, "source": ["E_I"], "target": ["EI"]},
        "k4": {"k": kinactf*1, "source": ["I"], "target": ["-1EI"]},
        "k5": {"k": kinactf*1, "source": ["EI"], "target": ["2E_I","I","E"]},
    },
    "dofit": False,
}

gk.utils.plotting.assign_colors_to_species(schemes,saturation_range=(0.5,0.8),lightness_range=(0.4,0.5),overwrite_existing=False,seed=1)

# Hardcoded 3S systems 
class ThreeStateVani():
    def __init__(self,Ki,koff,kinactf,kinactb,conc0Arr):
        self.Ki = Ki
        self.koff = koff
        self.kon = self.koff / Ki
        self.kinactf=kinactf
        self.kinactb=kinactb
        self.conc0Arr = conc0Arr
        self.sol=None

    @staticmethod
    def ode(t,concArr, params):
        kon,koff,kinactf,kinactb=params
        I, E, E_I, EI  = concArr
        dIdt = koff*E_I - kon*I*E
        dEdt = koff*E_I - kon*I*E 
        dE_Idt = kon*I*E - koff*E_I - kinactf*E_I + kinactb*EI
        dEIdt = kinactf*E_I - kinactb*EI
        return [dIdt, dEdt, dE_Idt, dEIdt]
        
    def solve(self, t):
        t_span = (t[0], t[-1])
        params = (self.kon, self.koff, self.kinactf, self.kinactb)
        solution = solve_ivp(self.ode, t_span, self.conc0Arr, t_eval=t, args=(params,), method='BDF', rtol=1e-6, atol=1e-8)
        return solution

class ThreeStateMod1dot1():
    def __init__(self,Ki,koff,kinactf,kinactb,conc0Arr):
        self.Ki = Ki
        self.koff = koff
        self.kon = self.koff / Ki
        self.kinactf=kinactf
        self.kinactb=kinactb
        self.conc0Arr = conc0Arr
        self.sol=None

    @staticmethod
    def ode(t,concArr, params):
        kon,koff,kinactf,kinactb=params
        I, E, E_I, EI  = concArr
        dIdt = 2*koff*E_I - 2*kon*(I**2)*E
        dEdt = koff*E_I - kon*(I**2)*E 
        dE_Idt = kon*(I**2)*E - koff*E_I - 3*kinactf*(E_I**3) + 3*kinactb*(EI**7)
        dEIdt = 7*kinactf*(E_I**3) - 7*kinactb*(EI**7)
        return [dIdt, dEdt, dE_Idt, dEIdt]
        
    def solve(self, t):
        t_span = (t[0], t[-1])
        params = (self.kon, self.koff, self.kinactf, self.kinactb)
        solution = solve_ivp(self.ode, t_span, self.conc0Arr, t_eval=t, args=(params,), method='BDF', rtol=1e-6, atol=1e-8)
        return solution

class ThreeStateMod1dot2():
    def __init__(self,Ki,koff,kinactf,kinactb,conc0Arr):
        self.Ki = Ki
        self.koff = koff
        self.kon = self.koff / Ki
        self.kinactf=kinactf
        self.kinactb=kinactb
        self.conc0Arr = conc0Arr
        self.sol=None

    @staticmethod
    def ode(t,concArr, params):
        kon,koff,kinactf,kinactb=params
        I, E, E_I, EI  = concArr
        dIdt = koff*E_I + koff*E_I - kon*I*I*E - kon*I*I*E
        dEdt = koff*E_I - kon*I*I*E 
        dE_Idt = kon*I*I*E - koff*E_I - kinactf*E_I*E_I*E_I - kinactf*E_I*E_I*E_I - kinactf*E_I*E_I*E_I + kinactb*(EI**7) + kinactb*(EI**7) + kinactb*(EI**7)
        dEIdt = 7*kinactf*E_I*E_I*E_I - 7*kinactb*(EI**7)
        return [dIdt, dEdt, dE_Idt, dEIdt]
        
    def solve(self, t):
        t_span = (t[0], t[-1])
        params = (self.kon, self.koff, self.kinactf, self.kinactb)
        solution = solve_ivp(self.ode, t_span, self.conc0Arr, t_eval=t, args=(params,), method='BDF', rtol=1e-6, atol=1e-8)
        return solution

class ThreeStateMod2dot1():
    def __init__(self,Ki,koff,kinactf,kinactb,conc0Arr):
        self.Ki = Ki
        self.koff = koff
        self.kon = self.koff / Ki
        self.kinactf=kinactf
        self.kinactb=kinactb
        self.conc0Arr = conc0Arr
        self.sol=None

    @staticmethod
    def ode(t,concArr, params):
        kon,koff,kinactf,kinactb=params
        I, E, E_I, EI  = concArr
        dIdt = koff*E_I*E - kon*I*(E**2) 
        dEdt = 2*koff*E_I*E - 2*kon*I*(E**2) + kon*I*(E**2) - koff*E_I*E
        dE_Idt = kon*I*(E**2) - koff*E_I*E - kinactf*E_I + kinactb*EI
        dEIdt = kinactf*E_I - kinactb*EI
        return [dIdt, dEdt, dE_Idt, dEIdt]
        
    def solve(self, t):
        t_span = (t[0], t[-1])
        params = (self.kon, self.koff, self.kinactf, self.kinactb)
        solution = solve_ivp(self.ode, t_span, self.conc0Arr, t_eval=t, args=(params,), method='BDF', rtol=1e-6, atol=1e-8)
        return solution
    
class ThreeStateOsc1():
    def __init__(self,Ki,koff,kinactf,kinactb,conc0Arr):
        self.Ki = Ki
        self.koff = koff
        self.kon = self.koff / Ki
        self.kinactf=kinactf
        self.kinactb=kinactb
        self.conc0Arr = conc0Arr
        self.conc0Arr[0] = self.conc0Arr[0]*1e-2 #I starts at 1
        self.sol=None

    @staticmethod
    def ode(t,concArr, params):
        kon,koff,kinactf,kinactb=params
        kon = kon*8e1
        koff = koff*5e2
        kinactf = kinactf*1
        
        I, E, E_I, EI  = concArr

        dIdt =  - 0 * kon * E * I**0\
                + 0 * koff * E_I\
                - kinactf * I\
                + kinactf * EI
        
        dEdt =  - kon * E * I**0\
                + -1 * koff * E_I\
                + kinactf * EI
        
        dE_Idt =  + kon * E * I**0\
                - koff * E_I\
                + koff * E_I\
                - kinactf * E_I\
                + 2 * kinactf * EI
        
        dEIdt =  + kinactf * E_I\
                + -1 * kinactf * I\
                - kinactf * EI
        
        return [dIdt, dEdt, dE_Idt, dEIdt]
        
    def solve(self, t):
        t_span = (t[0], t[-1])
        params = (self.kon, self.koff, self.kinactf, self.kinactb)
        solution = solve_ivp(self.ode, t_span, self.conc0Arr, t_eval=t, args=(params,), method='BDF', rtol=1e-6, atol=1e-8)
        return solution

# Old NState functions for making sure new versions produce the same results
def _ode_old(system, t, concentrations):
    """
    Compute the derivative of concentrations with respect to time.
    Can directly handle non-linear reactions (eg stoich coeff > 1)
    """ 
    ode_arr = np.zeros_like(concentrations)
    for tr in system.transitions.values():
        rate_constant = tr.k
        rate = rate_constant * np.prod([concentrations[system.species[sp_name].index] ** coeff for sp_name,coeff in tr.source])
        # Iterating through like this is beneficial because it captures stoichiometry that is evident in the list rather than coefficient 
            # (eg "source": ["E","E", "I"] is equal to "source": ["2E", "I"])
        for sp_name,coeff in tr.source:
            ode_arr[system.species[sp_name].index] -= coeff * rate
        for sp_name,coeff in tr.target:
            ode_arr[system.species[sp_name].index] += coeff * rate
    return ode_arr

def _solve_odes_old(system, t, method='BDF', rtol=1e-6, atol=1e-8, output_raw=False):
    """
    Solve the ODEs for the system.
    """
    conc0 = np.array([np.atleast_1d(sp.y0) for _, sp in system.species.items()])
    t_span = (t[0], t[-1])
    system.log_odes()

    try:
        solution = solve_ivp(
            fun=lambda t, conc: _ode_old(system, t, conc),
            t_span=t_span, y0=conc0, t_eval=t, method=method, rtol=rtol, atol=atol,
        )
        if not solution.success:
            raise RuntimeError("ODE solver failed: " + solution.message)

        system.t = t
        for name,data in system.species.items():
            data.simout["y"] = solution.y[data.index]

        system.logger.info("ODEs solved successfully.")
        if output_raw:
            return solution
        else:
            return
    except Exception as e:
        system.logger.error(f"Error in solving ODEs: {e}")
        raise

# Solve gekim systems
def solve_gekim_systems(schemes,plot_timecourse=False) -> None: 
    for name,scheme in schemes.items():
        system = gk.schemes.NState(scheme,quiet=True)
        #sol = _solve_odes_old(system,t,output_raw=True) #testing with old (nonvectorized) version
        system.set_simulator(gk.simulators.ODESolver)
        sol = system.simulator.simulate(t_eval=t,output_raw=True,atol=1e-8,rtol=1e-6)
        final_state = system.species["EI"].simout["y"]
        all_bound = system.sum_species_simout(blacklist=["E","I"])
        if scheme.get("dofit",True):
            fit_output = ii.kobs_uplim_fit_to_occ_final_wrt_t(
                t,final_state,nondefault_params={"Etot":{"fix":concE0}})
        else:
            fit_output = None
        schemes[name]["fit_output"] = fit_output
        schemes[name]["sol"] = sol
        
        if plot_timecourse:
            fig = plt.figure(figsize=(5, 3))

            # Time course of species 
            for species, sp_data in system.species.items():
                if species == "I":
                    continue
                plt.plot(t, sp_data.simout["y"], label=sp_data.label,color=sp_data.color)

            # All bound states
            all_bound = system.sum_species_simout(blacklist=["E","I"])
            plt.plot(t, all_bound,label='All Bound States',color="grey")

            # Fitted data
            plt.plot(t,  fit_output.y_fit,label=r"New Fit: $k_{\text{obs}}$ = "+
                    str(gk.utils.helpers.round_sig(fit_output.fitted_params["kobs"],3))+r" $\text{s}^{-1}$",ls='--', color="black")

            plt.xlabel('Time (s)')
            plt.ylabel('Concentration (nM)')
            plt.legend(frameon=False)
            plt.show()

# Solve hardcoded systems
def solve_hardcoded_system(system,t,concE0,dofit=True):
    sol = system.solve(t)
    final_state = sol.y[-1]
    if dofit:
        fit_output = ii.kobs_uplim_fit_to_occ_final_wrt_t(
            t,final_state,nondefault_params={"Etot":{"fix":concE0}})
    else: 
        fit_output = None
    system_dict = {
        "system": system,
        "sol": sol,
        "final_state": final_state,
        "fit_output": fit_output
    }
    return system_dict

def compare_systems(sys1_label,sys1_dict, sys2_label,sys2_dict, t, didfits=True) -> None:
    bad = False
    if np.allclose(sys1_dict["sol"].y, sys2_dict["sol"].y,rtol=1e-6):
        if didfits:
            if gk.utils.helpers.compare_dictionaries(sys1_dict["fit_output"].fitted_params,sys2_dict["fit_output"].fitted_params,rel_tol=1e-6,abs_tol=1e-8):
                print(f"GOOD\tsol and fit match: {sys1_label} and {sys2_label}","\n")
            else:
                print(f"BAD\tfit mismatch: {sys1_label} and {sys2_label}")
                print(f"{sys1_label}:\n",sys1_dict["fit_output"].fitted_params)
                print(f"{sys2_label}:\n",sys2_dict["fit_output"].fitted_params)
                print("\n")
                bad = True
        else:
            print(f"GOOD\tsol match: {sys1_label} and {sys2_label}","\n")
    else:
        print(f"BAD\tsol mismatch: {sys1_label} and {sys2_label}")
        print(f"{sys1_label}:\n",sys1_dict["sol"],"\n")
        print(f"{sys2_label}:\n",sys2_dict["sol"],"\n")
        print("\n")
        bad = True
    
    if bad:
        fig, axs = plt.subplots(2, 1, figsize=(8, 6))

        #axs[0].plot(t, sys1_dict["sol"][0], label="I")
        axs[0].plot(t, sys1_dict["sol"].y[1], label="E")
        axs[0].plot(t, sys1_dict["sol"].y[2], label="E_I")
        axs[0].plot(t, sys1_dict["sol"].y[3], label="EI")
        axs[0].set_xlabel("Time (s)")
        axs[0].set_ylabel("Concentration (nM)")
        axs[0].set_title(sys1_label)
        axs[0].legend()

        #axs[1].plot(t, sys2_dict["sol"][0], label="I")
        axs[1].plot(t, sys2_dict["sol"].y[1], label="E")
        axs[1].plot(t, sys2_dict["sol"].y[2], label="E_I")
        axs[1].plot(t, sys2_dict["sol"].y[3], label="EI")
        axs[1].set_xlabel("Time (s)")
        axs[1].set_ylabel("Concentration (nM)")
        axs[1].set_title(sys2_label)
        axs[1].legend()

        plt.tight_layout()
        plt.show()



def main() -> None:
    # Solve gekim systems
    solve_gekim_systems(schemes,plot_timecourse=PLOT)

    # Solve hardcoded systems
    conc0Arr = [concI0,concE0,0,0]
    hardcoded_vani = ThreeStateVani(Kd,koff,kinactf,kinactb,conc0Arr)
    vani_dict = solve_hardcoded_system(hardcoded_vani,t,concE0)
    hardcoded_mod1dot1 = ThreeStateMod1dot1(Kd,koff,kinactf,kinactb,conc0Arr)
    mod1dot1_dict = solve_hardcoded_system(hardcoded_mod1dot1,t,concE0)
    hardcoded_mod1dot2 = ThreeStateMod1dot2(Kd,koff,kinactf,kinactb,conc0Arr)
    mod1dot2_dict = solve_hardcoded_system(hardcoded_mod1dot2,t,concE0)
    hardcoded_mod2dot1 = ThreeStateMod2dot1(Kd,koff,kinactf,kinactb,conc0Arr)
    mod2dot1_dict = solve_hardcoded_system(hardcoded_mod2dot1,t,concE0)
    hardcoded_osc1 = ThreeStateOsc1(Kd,koff,kinactf,kinactb,conc0Arr)
    osc1_dict = solve_hardcoded_system(hardcoded_osc1,t,concE0,dofit=False)

    # Compare systems
    compare_systems("vani hardcoded", vani_dict, "vani gekim", schemes["3S_vani"], t)
    compare_systems("mod1.1 hardcoded", mod1dot1_dict, "mod1.1 gekim", schemes["3S_mod1.1"], t)
    compare_systems("mod1.2 hardcoded", mod1dot2_dict, "mod1.2 gekim", schemes["3S_mod1.2"], t)
    compare_systems("mod1.1 gekim", schemes["3S_mod1.1"], "mod1.2 gekim", schemes["3S_mod1.2"], t)
    compare_systems("mod2.1 hardcoded", mod2dot1_dict, "mod2.1 gekim", schemes["3S_mod2.1"], t)
    compare_systems("osc1 hardcoded", osc1_dict, "osc1 gekim", schemes["3S_osc1"], t,didfits=False)

if __name__ == "__main__":
    main()



    


import numpy as np
from .base_simulator import BaseSimulator

    #TODO: add an option called "continue" which takes an integer which points to the index of the run that its continuing from
    #TODO: precompile (repeatable) algo in c or fortran
    #TODO: S2.alt1 breaks this from negative rates somehow
    #TODO: running prob? and test more
    #TODO: type hints on algo
    #TODO: 2S I goes neg
    #TODO: warn if transition is not linear

class Gillespie(BaseSimulator):
    """
    Gillespie's algorithm for stochastic simulation.
    Does not work if any transitions are > (pseudo-)first order.  
    """
    def setup(self):
        pass

    def _process_simouts(self, simouts, y0_mat_len):
        if y0_mat_len == 1:
            simout = simouts[0]
            self.system.simout["t"] = simout['t']
            self.system.simout["prob_dist"] = simout['prob_dist']
            for sp_name, sp_data in self.system.species.items():
                sp_data.simout["prob_dist"] = simout['prob_dist'][:, sp_data.index]
        else:
            self.simout["t"] = [simout['t'] for simout in simouts]
            self.simout["prob_dist"] = [simout['prob_dist'] for simout in simouts]
            for sp_name, sp_data in self.system.species.items():
                sp_data.simout["prob_dist"] = [simout['prob_dist'][:, sp_data.index] for simout in simouts]
    
    def simulate(self, t_max, num_replicates=1, output_times=None, output_raw=False,**kwargs):
        """
        Run a Gillespie simulation on the system.
        
        Parameters:
        ----------
        t_max : float
            The maximum time to simulate.
        num_replicates : int, optional
            The number of replicates to run. Default is 1.
        output_times : array-like, optional
            The times at which to output the state of the system. Default is None.
        output_raw : bool, optional
            Whether to return the raw simulation output. Default is False.
        **kwargs : dict, optional
            Additional keyword arguments to pass to the simulation.

        Returns:
        -------
        dict or list of dict
            The simulation output. If `output_raw` is True, returns a dictionary or a list of dictionaries containing the simulation output for each replicate. If `output_raw` is False, returns None.

        """
        y0_mat = self._make_y0_mat()
        y0_mat_len = len(y0_mat)
        raw_simouts = [self.simulate_single(t_max, y0, num_replicates, output_times,**kwargs) for y0 in y0_mat]
        self._process_simouts(raw_simouts, y0_mat_len)
        if output_raw:
            if y0_mat_len == 1:
                self.system.log.info("Returning raw simulation output.\n")
                return raw_simouts[0]
            self.system.log.info("Returning list of raw simulation outputs.\n")        
            return raw_simouts
        else:
            self.system.log.info("Not returning raw simulation output. Use output_raw=True to return raw data.\n")
            return

    def simulate_single(self, t_max, conc0, num_replicates, output_times, max_iter=1000):
        simouts = [self._simulate_replicate(t_max, conc0, output_times, max_iter) for _ in range(num_replicates)]
        return self._aggregate_replicate_data(simouts)

    def _simulate_replicate(self, t_max, conc0, output_times, max_iter):
        times, states = [0], [conc0]

        while times[-1] < t_max and len(times) < max_iter:
            rates = self._calculate_transition_rates(states[-1])
            total_rate = np.sum(rates)
            if total_rate == 0:
                break
            time_step = np.random.exponential(1 / total_rate)
            if (new_time := times[-1] + time_step) > t_max:
                break
            times.append(new_time)
            chosen_transition = np.random.choice(len(rates), p=rates / total_rate)
            transitions = list(self.system.transitions.values())
            states.append(self._apply_transition(states[-1], transitions[chosen_transition]))

        return {'t': np.array(times), 'state': np.array(states)}

    def _aggregate_replicate_data(self, replicates):
        t_all = np.concatenate([rep['t'] for rep in replicates])
        t_edges = np.unique(t_all)
        prob_dist = np.mean([self._collect_states_at_times(rep['t'], rep['state'], t_edges) for rep in replicates], axis=0)
        return {'t': t_edges, 'prob_dist': prob_dist}

    def _collect_states_at_times(self, times, states, output_times):
        idxs = np.searchsorted(times, output_times, side='right') - 1
        idxs[idxs < 0] = 0
        return states[idxs]

    def _calculate_transition_rates(self, state):
        rates = []
        for tr in self.system.transitions.values():
            rate = tr.k * np.prod([state[self.system.species[sp_name].index] ** coeff for sp_name, coeff in tr.source])
            rates.append(rate)
        return np.array(rates)

    def _apply_transition(self, current_state, transition):
        new_state = np.array(current_state)
        for sp_name, coeff in transition.source:
            new_state[self.system.species[sp_name].index] -= coeff
        for sp_name, coeff in transition.target:
            new_state[self.system.species[sp_name].index] += coeff
        return new_state


import numpy as np
import math
from itertools import product


def rate_pair_from_P(int_OOM,P_B) -> tuple:
    """
    Provides rates between two states that are at rapid equilibrium and 
        therefore approximated by the population distribution. 

    Parameters:
    ----------
    int_OOM : int or float
        The order of magnitude of the rates.
    P_B : float
        The probability of the state B, i.e., kAB/(kBA+kAB) or [B]/([A]+[B]) for A<-->B.

    Returns:
    -------
    kAB : float
        The rate constant from A to B.
    kBA : float
        The rate constant from B to A.
    """
    kAB=P_B*10**(float(int_OOM)+1)
    kBA=10**(float(int_OOM)+1)-kAB
    return kAB,kBA

def integerable_float(num):
    """
    Returns a float as an integer if it is an integer, otherwise returns the float.
    """
    if num.is_integer():
        return int(num)
    else:
        return num
    
def round_sig(num: float, sig_figs: int = 3, autoformat=True):
    """
    Round up using significant figures.

    Parameters:
    ----------
    num : float
        The number to be rounded.
    sig_figs : int
        The number of significant figures to round to.
    autoformat : bool, optional
        If True (default), formats the result into scientific notation if the order of magnitude 
        is greater than +/- 3.

    Returns:
    -------
    float or str
        The rounded number. If autoformat is True and the result is in scientific notation, 
        it is returned as a string.
    """
    if num == 0:
        return 0.0

    order_of_magnitude = math.floor(math.log10(abs(num)))
    shift = sig_figs - 1 - order_of_magnitude

    # Scale the number, round it, and scale it back
    scaled_num = num * (10 ** shift)
    rounded_scaled_num = math.floor(scaled_num + 0.5)
    result = rounded_scaled_num / (10 ** shift)

    # Handling scientific notation conditionally
    if autoformat and (order_of_magnitude >= 4 or order_of_magnitude <= -4):
        return format(result, '.{0}e'.format(sig_figs - 1))
    else:
        return result

def update_dict_with_subset(defaults: dict, updates: dict):
    """
    Recursively updates the default dictionary with values from the update dictionary,
    ensuring that only the keys present in the defaults are updated.

    Parameters
    ----------
    defaults : dict
        The default dictionary containing all allowed keys with their default values.
    updates : dict
        The update dictionary containing keys to update in the defaults dictionary.

    Returns
    -------
    dict
        The updated dictionary.
    """

    for key, update_value in updates.items():
        if key in defaults:
            # If both the default and update values are dictionaries, recurse
            if isinstance(defaults[key], dict) and isinstance(update_value, dict):
                defaults[key] = update_dict_with_subset(defaults[key], update_value)
            else:
                defaults[key] = update_value

    return defaults

def compare_dictionaries(dict1, dict2, rel_tol=1e-9, abs_tol=0.0):
    """
    Compare two dictionaries recursively and check if their values are approximately equal.

    Parameters
    ----------
    dict1 : dict
        The first dictionary to compare.
    dict2 : dict
        The second dictionary to compare.
    rel_tol : float, optional
        The relative tolerance for comparing floating-point values. Default is 1e-9.
    abs_tol : float, optional
        The absolute tolerance for comparing floating-point values. Default is 0.0.

    Returns
    -------
    bool
        True if the dictionaries are approximately equal, False otherwise.
    """
    
    if dict1.keys() != dict2.keys():
        return False

    for key in dict1:
        value1, value2 = dict1[key], dict2[key]
        if isinstance(value1, dict) and isinstance(value2, dict):
            # Recursive call for nested dictionaries
            if not compare_dictionaries(value1, value2, rel_tol, abs_tol):
                return False
        else:
            arr1, arr2 = np.array(value1), np.array(value2)
            if not np.allclose(arr1, arr2, rtol=rel_tol, atol=abs_tol):
                return False
    return True



def generate_dictval_combinations(input_dict: dict):
    """
    Generate all combinations of values in a dictionary and return them as a dictionary
    with the original keys and numpy arrays of corresponding values.

    Parameters
    ----------
    input_dict : dict
        Keys are parameter names and values are numpy arrays.

    Returns
    -------
    dict
        Dictionary with the original keys and numpy arrays of corresponding value
    """

    names = list(input_dict.keys())
    values = list(input_dict.values())

    combinations = list(product(*values))
    combinations_array = np.array(combinations).T

    comb_dict = {name: combinations_array[i] for i, name in enumerate(names)}
    return comb_dict

def arr2float(value: np.ndarray):
    """
    Convert an size-1 array to float.
    """
    if isinstance(value,np.ndarray):
        if value.size == 1:
            return float(value[0])
        else:
            raise ValueError("Array is not size 1.")
    else:
        return value


import numpy as np
from lmfit import Parameters
from lmfit.model import ModelResult
from typing import Union

from ...utils.fitting import general_fit, merge_params

#TODO: normalize for fit for x axis as well

def res_time():
    raise NotImplementedError()

def dose_response(dose: np.ndarray,Khalf: float,n=1,uplim=1,nonspecific_m=0): 
    '''
    Calculates the Hill equation for a dose-response curve.

    Parameters
    ----------
    dose : np.ndarray
        Array of input concentrations of the ligand.
    Khalf : float
        The dose required for half output response.
    n : float, optional
        Hill coefficient, default is 1.
    uplim : float, optional
        The upper limit of the response, default is 1, ie 100%.
    nonspecific_m : float, optional
        The slope of the nonspecific term, default is 0.

    Returns
    -------
    np.ndarray
        The fraction of the responding population.
    '''
    return uplim / (1+(Khalf/dose)**n) + nonspecific_m*dose

def dose_response_fit(dose: np.ndarray, response: np.ndarray, nondefault_params: Union[dict,Parameters] = None, xlim: tuple = None, 
                        weights_kde=False, weights: np.ndarray = None, verbosity=2, **kwargs) -> ModelResult:
    """
    Fit parameters (Khalf, kinact, n) to response with respect to dose using 
    a structured dictionary for parameters.

    Parameters
    ----------
    dose : np.ndarray
        Array of input concentrations of the ligand.
    response : np.ndarray
        Array of the fraction of the responding population.
    nondefault_params : dict or Parameters, optional
        A structured dictionary of parameters with 'value','vary', and 'bound' keys or a lmfit.Parameters object.
        Defaults:
        ```python
        default_params.add('Khalf', value=100, vary=True, min=0, max=np.inf)
        default_params.add('uplim', value=1, vary=True)
        default_params.add('n', value=1, vary=True, min=0, max=np.inf)
        default_params.add('nonspecific_m', value=0, vary=False)
        ```
        Example dict of nondefaults:
        ```python
        nondefault_params = {
            "n": {"vary": False},    
        }
        ```
    xlim : tuple, optional
        Limits for the time points considered in the fit (min_t, max_t).
    weights_kde : bool, optional
        If True, calculate the density of the x-values and use the normalized reciprocol as weights. Similar to 1/sigma for scipy.curve_fit.
        Helps distribute weight over unevenly-spaced points. Default is False.
    weights : np.ndarray, optional
        weights parameter for fitting. This argument is overridden if weights_kde=True. Default is None.
    verbosity : int, optional
        0: print nothing. 1: print upon bad fit. 2: print always. Default is 2.
    kwargs : dict, optional
        Additional keyword arguments to pass to the lmfit Model.fit function.

    Returns
    -------
    lmfit.ModelResult
        The result of the fitting operation from lmfit.
    """
    default_params = Parameters()
    default_params.add('Khalf', value=100, vary=True, min=0, max=np.inf)
    default_params.add('uplim', value=1, vary=True)
    default_params.add('n', value=1, vary=True, min=0, max=np.inf)
    default_params.add('nonspecific_m', value=0, vary=False)

    lm_params = merge_params(default_params, nondefault_params)
    return general_fit(dose_response, dose, response, lm_params, xlim=xlim, weights_kde=weights_kde, weights=weights, verbosity=verbosity, **kwargs)

class Params:
    """
    Common place for parameters found in general binding Khalfnetics literature.
    """
    @staticmethod
    def Kd(kon, koff):
        """
        Kd (i.e. dissociation constant) calculation
        
        Parameters
        ----------
        kon : float
            On-rate constant (CONC^-1*TIME^-1)
        koff : float
            Off-rate constant (TIME^-1)
        
        Returns
        -------
        float
            The calculated dissociation constant (Kd)
        """
        return koff / kon
    
    @staticmethod
    def Keq(kon, koff):
        """
        Keq (i.e. equilibrium constant) calculation
        
        Parameters
        ----------
        kon : float
            On-rate constant (CONC^-1*TIME^-1)
        koff : float
            Off-rate constant (TIME^-1)
        
        Returns
        -------
        float
            The calculated equilibrium constant (Keq)
        """
        return kon / koff

    



import numpy as np
import re
import copy
from sympy import symbols
from collections import defaultdict
from typing import Union
from ..utils.helpers import integerable_float
from ..utils.logging import Logger
#todo: dataclasses for species transitions and paths 
class Species:
    def __init__(self, name: str, y0: Union[np.ndarray,float], label=None, color=None):
            """
            Initialize a species object.

            Parameters
            ----------
            name : str
                Name of the species.
            y0 : Union[np.ndarray,float]
                Initial concentration of the species.
                Array Example: `{"Ligand":np.linspace(1,1500,100)}` for a Michaelis-Menten ligand concentration scan.
            label : str, optional
                Useful for plotting. Will default to NAME.
            color : str, optional
                Useful for plotting. Best added by ..utils.Plotting.assign_colors_to_species().

            """
            self.name = name
            self.y0 = np.array([y0]) if np.isscalar(y0) else np.array(y0)
            self.label = label or name
            self.index = None  # added by NState
            self.color = color
            self.sym = symbols(name)
            self.simin = None  # added by simulator
            self.simout = None  # added by simulator

    def __repr__(self):
        return f"{self.name} (Initial Concentration: {self.y0}, Label: {self.label})"

class Transition:
    def __init__(self, name: str, k, source: list, target: list, label=None, index=None):
        """
        Parameters
        ----------
        name : str
            Name of the rate constant.
        k : float
            Value of the rate constant.
        source : list
            List of (SPECIES, COEFF) tuples or "{COEFF}{SPECIES}" strings.
        target : list
            List of (SPECIES, COEFF) tuples or "{COEFF}{SPECIES}" strings.
        label : str, optional
            Could be useful for plotting. Will default to NAME.
        """
        self.name = name # should be the name of the rate constant for all intents and purposes, eg "kon"
        self.k = k
        self.source = Transition._format_state(source,"source")  # List of (SPECIES, COEFF) tuples or "{COEFF}{SPECIES}" strings
        self.target = Transition._format_state(target,"target")  # List of (SPECIES, COEFF) tuples or "{COEFF}{SPECIES}" strings
        self.label = label or name
        self.index = index
        self.sym = symbols(name)
        self.simin = None # added by simulator
        self.simout = None # added by simulator
        self.linear = self.is_linear()
        

    def __repr__(self):
        source_str = ' + '.join([f"{coeff}*{sp}" for sp, coeff in self.source])
        target_str = ' + '.join([f"{coeff}*{sp}" for sp, coeff in self.target])
        return f"{self.name} ({self.k}): {source_str} -> {target_str}"

    def is_linear(self):
        """
        Check if a transition is linear.

        Returns
        -------
        bool
            True if the transition is linear, False otherwise.

        Notes
        -----
        A first-order reaction must have exactly one source species (with a stoichiometric coefficient of 1).
        """
        return len(self.source) == 1 and self.source[0][1] == 1

    @staticmethod
    def _parse_species_string(species_str):
        """
        Extract coefficient and species name from species string.

        Parameters
        ----------
        species_str : str
            A species string, e.g., '2A'.

        Returns
        -------
        tuple
            A tuple of species name (str) and stoichiometric coefficient (int).
        """
        match = re.match(r"(-?\d*\.?\d*)(\D.*)", species_str)
        if match and match.groups()[0]:
            coeff = match.groups()[0]
            if coeff == '-':
                coeff = -1
            coeff = integerable_float(float(coeff))
        else:
            coeff = 1
        name = match.groups()[1] if match else species_str
        return name,coeff

    @staticmethod            
    def _format_state(state,direction=None):
        """
        Format a transition by extracting and combining coefficients and species names.
        Is idempotent.

        Parameters
        ----------
        state : list
            State descriptor. List of (SPECIES, COEFF) tuples or "{COEFF}{SPECIES}" strings.
        direction : str, optional
            Direction of the transition. Default is None. Can be "source" or "target".

        Returns
        -------
        list
            List of (SPECIES, COEFF) tuples.

        Raises
        ------
        ValueError
            If the transition or species tuples are invalid.
        """
        parsed_species = {}
        for sp in state:
            if isinstance(sp, str):
                name, coeff = Transition._parse_species_string(sp)
            elif isinstance(sp, tuple):
                if len(sp) == 2:
                    if isinstance(sp[0], str) and isinstance(sp[1], (int, float)):
                        name, coeff = sp
                    elif isinstance(sp[1], str) and isinstance(sp[2], (int, float)):
                        coeff, name = sp
                    else:
                        raise ValueError(f"Invalid species tuple '{sp}' in transition '{state}'.")
                else:
                    raise ValueError(f"Invalid species tuple '{sp}' in transition '{state}'.")
            else:
                raise ValueError(f"Invalid species '{sp}' in transition '{state}'.")
            if direction == "source" and coeff < 0:
                raise ValueError(f"Negative coefficient '{coeff}' in source of transition '{state}'.")
            if name in parsed_species:
                parsed_species[name] += coeff # combine coeffs
            else:
                parsed_species[name] = coeff
        state = [(name, coeff) for name, coeff in parsed_species.items()]
        return state
    
class Path:
    """
    Represents a path in a network of species transitions. 
    Is created by `NState.find_paths()`

    Attributes
    ----------
    species_path : list
        List of species objects representing the path.
    transitions_path : list
        List of transition objects representing the transitions along the path.
    probability : float
        Probability of the path relative to other paths from `species[0]` to `species[-1]`
    length : int
        Length of species_path.

    Methods
    -------
    __repr__()
        Returns a string representation of the species path.

    """

    def __init__(self, species_path, transitions_path, probability):
        self.species_path = species_path
        self.transitions_path = transitions_path
        self.probability = probability
        self.length = len(species_path)

    def __repr__(self):
        """
        Returns a string representation of the Path object.

        Returns
        -------
        str
            String representation of the Path object.

        """
        path_str = ' -> '.join(['+'.join([sp.name for sp in group]) if isinstance(group, list) else group.name for group in self.species_path])
        prob_fmt = "{:.2e}".format(self.probability)
        return f"Path(Length: {str(self.length).rjust(3)},\tProbability: {prob_fmt},\t{path_str})"

class NState:
    #TODO: add/remove_species and add/remove_transition methods
    #TODO: reinits if species or transitions are modified. let user able to define custom rate law funcs 
    #TODO: markovian, nonmarkovian, etc
    #TODO: add a function for fetching info. say get_simout("species_name")

    
    def __init__(self, config: dict, logfilename=None, quiet=False):
            """
            Initialize the NState class with configuration data. Can be any degree of nonlinearity.

            Parameters
            ----------
            config : dict
                Configuration containing species and transitions.
                Species should contain name, initial concentration, and label.
                Transitions should contain name, source-species, target-species, and k value.
            logfilename : str, optional
                Name of the log file (default is None).
            quiet : bool, optional
                Flag indicating whether to suppress log output (default is False).

            Raises
            ------
            ValueError
                If config is invalid.
            """
            self.log = Logger(quiet=quiet, logfilename=logfilename)

            self._validate_config(config)
            self.config = copy.deepcopy(config)
        
            self.species = {
                name: Species(
                    name=name,
                    y0=np.array([data["y0"]]) if np.isscalar(data["y0"]) else np.array(data["y0"]),
                    label=data.get('label', name),
                    color=data.get('color')
                ) for name, data in config['species'].items()
            }

            self.transitions = {
                name: Transition(
                    name=name,
                    k=data['k'],
                    source=data['source'],
                    target=data['target'],
                    label=data.get('label', name)
                ) for name, data in config['transitions'].items()
            }

            self.setup()
            self.simulator = None
            self.paths = None
            self.log.info(f"NState system initialized successfully.\n")

    def setup(self):
        """
        Reinitialize the system after adding transitions or species.
        This method should be called if you modify the scheme after initialization.
        WARNING: This will reinitialize everything except the logger and config.
        """
        #TODO: this needs testing. Make sure that its fine to not reinit the concentrations
        # Document the order of the species
        for idx, name in enumerate(self.species):
            self.species[name].index = idx
        self._validate_species()
            
        # Document the order of the transitions
        for idx, name in enumerate(self.transitions):
            self.transitions[name].index = idx

        return

    def _validate_config(self,config):
        if not 'species' in config or not 'transitions' in config:
            raise ValueError("Config must contain 'species' and 'transitions' keys.")
        return True

    def _validate_species(self):
        #TODO: use assign color 
        labels = set()
        for name, data in self.species.items():
            # Validate labels
            label = data.label
            if label in labels:
                self.log.error(f"Duplicate label '{label}' found for species '{name}'.")
                return False
            labels.add(label)
        return True
    
    def simulate(self, simulator=None, *args, **kwargs):
        """
        Simulate the system using the provided simulator.

        Parameters
        ----------
        simulator : class, optional
            The simulator class to use for the system. Unless using a custom simulator, 
            use the provided simulators in gekim.simulators.
        *args : tuple, optional
            Additional arguments to pass to the simulator.simulate()
        **kwargs : dict, optional
            Additional keyword arguments to pass to the simulator.simulate()
        
        Returns
        -------
        Returns NState if the simulator didn't return anything, else returns the output of the simulator.

        Notes
        -----
        The simulator is forced to ONLY take NState (self) as an argument for initialization. 
        If the simulator requires additional arguments, initialize the simulator in an extra step, like so:
        ```python
        system.simulator = simulator(system, *args, **kwargs)
        system.simulator.simulate(*args, **kwargs)
        ```

        or
                
        ```python
        system.set_simulator(simulator, *args, **kwargs)
        system.simulator.simulate(*args, **kwargs)
        ```
        """
        if not simulator:
            if not self.simulator:
                self.log.error("Simulator not set. Use as an argument in NState.simulate() or set an initialized simulator to NState.simulator")
                return 
            else:
                simout = self.simulator.simulate(*args, **kwargs)
                if simout:
                    return simout
        else:
            self.log.info(f"Simulating with {simulator.__name__}.\n")
            simulator = simulator(self)
            simout = simulator.simulate(*args, **kwargs)
            if simout:
                return simout
        return self

    def set_simulator(self, simulator, *args, **kwargs) -> None:
        """
        Sets and initializes the simulator for the system. 

        Parameters
        ----------
        simulator : class
            The simulator class to use for the system. Unless using a custom simulator, 
            use the provided simulators in gekim.simulators.
        *args : tuple, optional
            Additional arguments to pass to the simulator for initialization.
        **kwargs : dict, optional
            Additional keyword arguments to pass to the simulator for initialization.

        Notes
        -----
        This method is not as good as just doing:
        ```python
        system.simulator = simulator(system)
        system.simulator.simulate(...)
        ```
        because IDE syntax and doc helpers may not pick up the new simulator attribute and simulate method.
        """
        self.simulator = simulator(self, *args, **kwargs)
        self.log.info(f"Simulator set to {simulator.__name__}.\n")
        self.log.info(f"Use system.simulator.simulate() or system.simulate() to run the simulation.\n")

    def sum_species_simout(self,whitelist:list=None,blacklist:list=None):
        """
        Sum the simout y-values of specified species.

        Parameters
        ----------
        whitelist : list, optional
            Names of species to include in the sum.
        blacklist : list, optional
            Names of species to exclude from the sum.

        Returns
        -------
        numpy.ndarray or None
            The sum of the simulated values. Returns None if the 
            simulated data is not found for any species.

        Raises
        ------
        ValueError
            If both whitelist and blacklist are provided.

        """
        if whitelist and blacklist:
            raise ValueError("Provide either a whitelist or a blacklist, not both.")

        species_names = self.species.keys()
        
        if whitelist:
            species_names = [name for name in whitelist if name in species_names]
        elif blacklist:
            species_names = [name for name in species_names if name not in blacklist]

        try:
            total_y = np.sum([self.species[name].simout["y"] for name in species_names], axis=0)
        except KeyError as e:
            self.log.error(f"Simulated data not found for species '{e}'.")
            return None
        return total_y

    def mat2sp_simout(self,matrix,key_name="y"):
        """
        Save species vectors from a concentration matrix to the respective 
        `species[NAME].simout[key_name]` dict based on `species[NAME].index`.
        
        Parameters
        ----------
        matrix : numpy.ndarray
            The concentration matrix containing the species vectors.
        key_name : str, optional
            The key name to use for saving the species vectors in the species dictionary (default is "y").

        Notes
        -----
        Useful for saving the output of a continuous solution to the species dictionary.
        Don't forget to save time, too, eg `system.simout["t_cont"] = t`
        """
        for _, sp_data in self.species.items():
            sp_data.simout[key_name] = matrix[sp_data.index]
        return

    def find_paths(self, start_species: Union[str,Species], end_species: Union[str,Species], only_linear_paths=True, 
                   prob_cutoff=1e-10, max_depth=20, log_paths=False, normalize_prob=True):
        """
        Find paths from start_species to end_species.

        Parameters
        ----------
        start_species : str or Species
            Name or object of the starting species.
        end_species : str or Species
            Name or object of the ending species.
        only_linear_paths : bool, optional
            Whether to only find linear paths (no backtracking or loops) (default is True).
        prob_cutoff : float, optional
            Cutoff probability to stop searching current path (default is 1e-10). This is before normalization of probabilities.
        max_depth : int, optional
            Maximum depth to limit the search (default is 20), ie max length of path - 1. 
        log_paths : bool, optional
            Whether to log the path strings found (default is False).

        Notes
        -------
        Saves a list of paths in `self.paths` sorted by probability.

        Probability may be misleading here due to the cutoffs and infinite possibilities of nonlinear paths. 
            
        Probability is calculated as the product of the transition probabilities, 
            which is the transition rate constant over the sum of available transition rate constants (markov chain-esque)
            
        """
        #TODO: use J_sym?
        #TODO: prob seems right, but why isnt it what is expected?
        #TODO: needs to be optimized, probably with multithreading. but since its main use is for finding linear systems, its fine
            #TODO: needs to be optimized in many ways, including algorithmic. Many wasted or repeat cycles  
        
        def get_transition_probability(transition, current_sp_name):
            k_sum = sum(tr.k for tr in self.transitions.values() if current_sp_name in [sp[0] for sp in tr.source])
            return transition.k / k_sum if k_sum > 0 else 0

        def dfs(current_sp_name, target_sp_name, visited_names, current_path, current_transitions, current_prob, depth):
            if current_prob < prob_cutoff or depth > max_depth:
                return
            
            if current_sp_name == target_sp_name:
                self.paths.append(Path(current_path[:], current_transitions[:], current_prob))
                return

            for transition in self.transitions.values():
                if current_sp_name in [sp[0] for sp in transition.source]:
                    next_species_list = [sp[0] for sp in transition.target]
                    if only_linear_paths and any(sp in visited_names for sp in next_species_list):
                        continue

                    for next_sp_name in next_species_list:
                        next_prob = current_prob * get_transition_probability(transition, current_sp_name)
                        #print(f"{current_sp_name} -> {next_sp_name} ({current_prob}->{next_prob}) by transition {transition.name}")
                        visited_names.add(next_sp_name)
                        current_path.append(self.species[next_sp_name])
                        current_transitions.append(transition)
                        dfs(next_sp_name, target_sp_name, visited_names, current_path, current_transitions, next_prob, depth + 1)
                        if only_linear_paths: #bandaid? did it work?
                            visited_names.remove(next_sp_name)
                        current_path.pop()
                        current_transitions.pop()

        # Input validation
        all_linear_tr = True
        for transition in self.transitions.values():
            if not transition.linear:
                all_linear_tr = False
                self.log.warning(f"Transition '{transition.name}' is not linear!")
        if not all_linear_tr:
            self.log.error("This method only uses TRANSITION.k to calculate probabilities, and expects single TRANSITION.source to contain only one species.\n" +
                           "If possible, make all transitions linear (e.g., with a pseudo-first-order approximation).\n")

        if isinstance(start_species, str):
            start_species = self.species[start_species]
        elif isinstance(start_species, Species):
            pass
        else:
            raise ValueError("start_species must be a string or Species object.")
    
        if isinstance(end_species, str):
            end_species = self.species[end_species]
        elif isinstance(end_species, Species):
            pass
        else:
            raise ValueError("end_species must be a string or Species object.")

        # Search
        self.paths = []
        dfs(start_species.name, end_species.name, {start_species.name}, [start_species], [], 1.0, 0) 

        # Normalize probabilities
        if normalize_prob:
            total_prob = sum(path.probability for path in self.paths)
            if total_prob > 0:
                for path in self.paths:
                    path.probability /= total_prob
    
        # Sort paths by probability
        self.paths.sort(key=lambda p: p.probability, reverse=True)

        
        if log_paths:
            self.log.info(f"\n{len(self.paths)} paths found from '{start_species.name}' to '{end_species.name}':")
            for path in self.paths:
                self.log.info(str(path))
        else:
            self.log.info(f"\n{len(self.paths)} paths found from '{start_species.name}' to '{end_species.name}'")
        
        return 

    def get_species_sets(self) -> dict:
        """
        Combine the probabilities of self.paths that contain the same set of species.
        Essentially a utility function if `NState.find_paths()` yields a ton of paths.

        Returns
        -------
        dict
            Dictionary with species sets as keys and combined probabilities as values.
        """
        paths = self.paths
        combined_paths = defaultdict(lambda: {"combined_probability": 0.0})

        for path in paths:
            species_set = frozenset(sp.name for sp in path.species_path)
            combined_paths[species_set]["combined_probability"] += path.probability

        # Sort combined paths by combined probability
        sorted_combined_paths = dict(sorted(combined_paths.items(), key=lambda item: item[1]['combined_probability'], reverse=True))

        self.log.info(f"\nSpecies sets and their combined probabilities (sorted):")
        for species_set, data in sorted_combined_paths.items():
            prob_fmt = "{:.2e}".format(data['combined_probability'])
            self.log.info(f"Combined P: {prob_fmt}, Species: {species_set}")

        return sorted_combined_paths



import numpy as np
from inspect import signature
from sympy import symbols, Matrix, lambdify
from typing import Union, Callable
from scipy.stats import gaussian_kde
from lmfit import Model, Parameters
from lmfit.model import ModelResult

from .helpers import arr2float


#TODO: scheme fitting, rate constant fitting
                
def general_fit(model_func: Callable, x: np.ndarray, y: np.ndarray, 
                params: Union[dict, Parameters], xlim: tuple = None, 
                weights_kde=False, weights: np.ndarray = None,
                verbosity=2, **kwargs) -> ModelResult:
    """
    General fitting function using lmfit.

    Parameters
    ----------
    model_func : callable
        The model function to fit. Should be of the form model_func(indep_var, *args, **kwargs).
    x : np.ndarray
        Array of independent variable.
    y : np.ndarray
        Array of observed data.
    params : dict
        Dictionary of parameters with 'fixed', 'guess', and 'bounds' keys.
    xlim : tuple, optional
        Limits for the x points considered in the fit (min_x, max_x).
    weights_kde : bool, optional
        If True, calculate the density of the x-values and use the normalized reciprocol as weights. Similar to 1/sigma for scipy.curve_fit.
        Helps distribute weight over unevenly-spaced points. Default is False.
    weights : np.ndarray, optional
        weights parameter for fitting. This argument is overridden if weights_kde=True. Default is None.
    verbosity : int, optional
        0: print nothing. 1: print upon bad fit. 2: print always. Default is 2.
    kwargs : dict, optional
        Additional keyword arguments to pass to the lmfit Model.fit function.

    Returns
    -------
    ModelResult
        The result of the fitting operation from lmfit.

    """
        
    if xlim:
        indices = (x >= xlim[0]) & (x <= xlim[1])
        x = x[indices]
        y = y[indices]

    if weights_kde:
        kde = gaussian_kde(x)
        weights = 1/kde(x)
        weights /= np.sum(weights)  # Normalize

    else:
        weights = weights

    if isinstance(params, Parameters):
        lm_params = params
    elif isinstance(params, dict):
        lm_params = Parameters()
        for name, info in params.items():
            if "vary" in info and not info["vary"]:
                lm_params.add(name, value=info["value"], vary=False)
            else:
                lm_params.add(name, value=info["value"], min=info["bounds"][0], max=info["bounds"][1])
    else:
        raise ValueError("params must be a lmfit.Parameters or dict instance.")

    # Dynamically handle the independent variable name so that model funcs aren't required to use "x"
    sig = signature(model_func)
    indep_var_name = list(sig.parameters.keys())[0]

    model = Model(model_func)
    model_result = model.fit(y, lm_params, **{indep_var_name: x}, weights=weights, **kwargs)

    if verbosity >= 1:
        bad_fit, message = detect_bad_fit(model_result)
        if bad_fit:
            print(f"Bad fit detected: {message}\n")

        if verbosity >= 2 or bad_fit:
            model_result.params.pretty_print();print('\n')


    return model_result

def merge_params(default_params: Parameters, nondefault_params: Union[dict, Parameters] = None) -> Parameters:
    """
    Merge default and nondefault parameters into a Parameters object.

    Parameters
    ----------
    default_params : Parameters
        The default parameters.
    nondefault_params : dict or Parameters, optional
        The nondefault parameters to override the defaults.

    Returns
    -------
    Parameters
        Merged Parameters object.
    """
    if nondefault_params is None:
        return default_params

    merged_params = default_params.copy()


    if isinstance(nondefault_params, Parameters):
        for name, param in nondefault_params.items():
            if name in merged_params:
                merged_params[name].set(
                    value=arr2float(param.value),
                    vary=param.vary,
                    min=param.min,
                    max=param.max,
                    expr=param.expr,
                    brute_step=param.brute_step,
                )
            else:
                merged_params.add(name,
                    value=arr2float(param.value),
                    vary=param.vary,
                    min=param.min,
                    max=param.max,
                    expr=param.expr,
                    brute_step=param.brute_step,
                )
    elif isinstance(nondefault_params, dict):
        for name, info in nondefault_params.items():
            if name in merged_params:
                merged_params[name].set(
                    value=arr2float(info.get("value", merged_params[name].value)),
                    vary=info.get("vary", merged_params[name].vary),
                    min=info.get("min", merged_params[name].min),
                    max=info.get("max", merged_params[name].max),
                    expr=info.get("expr", merged_params[name].expr),
                    brute_step=info.get("brute_step", merged_params[name].brute_step),
                )
            else:
                merged_params.add(name,
                    value=arr2float(info.get("value")),
                    vary=info.get("vary", True),
                    min=info.get("min", -np.inf),
                    max=info.get("max", np.inf),
                    expr=info.get("expr", None),
                    brute_step=info.get("brute_step", None),
                )
    else:
        raise ValueError("nondefault_params must be a dict or lmfit.Parameters object.")

    return merged_params

def chi_squared(observed_data: np.ndarray, fitted_data: np.ndarray, fitted_params: np.ndarray, variances: np.ndarray = None, reduced=False):
    """
    Calculate the chi-squared and optionally the reduced chi-squared statistics.

    Parameters
    ----------
    observed_data : np.ndarray
        The observed data points.
    fitted_data : np.ndarray
        The fitted data points obtained from curve 
    fitted_params : list or np.ndarray
        The optimal parameters obtained from curve 
    variances : np.ndarray, optional
        Variances of the observed data points. If None, assume constant variance.
    reduced : bool, optional
        If True, calculate and return the reduced chi-squared.

    Returns
    -------
    float
        The chi-squared or reduced chi-squared statistic.
    """

    if len(observed_data) != len(fitted_data):
        raise ValueError("Length of observed_data and fitted_data must be the same.")

    if len(fitted_params) == 0:
        raise ValueError("fitted_params cannot be empty.")

    residuals = observed_data - fitted_data
    chi_squared = np.sum((residuals**2) / variances) if variances is not None else np.sum(residuals**2)

    if reduced:
        degrees_of_freedom = len(observed_data) - len(fitted_params)
        if degrees_of_freedom <= 0:
            raise ValueError("Degrees of freedom must be positive.")
        return chi_squared / degrees_of_freedom

    return chi_squared

def _extract_fit_info(params):
    """Extracts initial guesses and bounds, and separates fixed and fitted parameters."""
    p0, bounds, param_order = [], ([], []), []
    fixed_params = {}
    for param, config in params.items():
        if config["fix"] is not None:
            fixed_params[param] = config["fix"]
        else:
            p0.append(config["guess"])
            bounds[0].append(config["bounds"][0])
            bounds[1].append(config["bounds"][1])
            param_order.append(param)
    return p0, bounds, param_order, fixed_params

class FitOutput:
    """
    Comprises the output of a fitting operation.

    Attributes
    ----------
    fitted_params : dict
        Parameters obtained from the fit, keyed by parameter name and zipped in the order indexed in pcov.
    pcov : array
        Covariance matrix of the fitted parameters.
    x : array
        Independent variable used for fitting.
    y_fit : array
        Data generated by the fitted params.
    y_obs : array
        Observed data used for fitting. May be normalized.
    reduced_chi_sq : float
        Reduced chi-squared value indicating goodness of fit.
    """
    def __init__(self, x, fitted_ydata, observed_ydata, fitted_params, pcov, reduced_chi_sq):
        self.fitted_params = fitted_params
        self.pcov = pcov
        self.x = x
        self.y_fit = fitted_ydata
        self.y_obs = observed_ydata
        self.reduced_chi_sq = reduced_chi_sq

def _prepare_output(x, fitted_ydata, observed_ydata, popt, pcov, param_order):
    """
    Prepare the fitting output as an instance of the FitOutput class.
    """
    fitted_params = dict(zip(param_order, popt))
    reduced_chi_sq = chi_squared(observed_ydata, fitted_ydata, popt, np.var(observed_ydata), reduced=True)
    return FitOutput(x, fitted_ydata, observed_ydata, fitted_params, pcov, reduced_chi_sq)

def generate_jacobian_func(fitting_adapter, param_order):
    '''
    Generate and store the Jacobian function for a given model function.
    Makes curve_fit slower for simple functions.

    Parameters
    ----------
    fitting_adapter : callable
        The fitting adapter function to fit. Should be of the form fitting_adapter(x, *params).
    param_order : list of str
        List of parameter names for the fitting adapter function.

    Returns
    -------
    jac_func : Callable
        A function that calculates the Jacobian matrix given the parameters.
    '''
    # Inspect the model function to get parameter names
    sig = signature(fitting_adapter)
    x_sym = list(sig.parameters.keys())[0]          # Independent variable
    x_sym = symbols(x_sym)
    param_syms = symbols(param_order)
    
    model_expr = fitting_adapter(x_sym, *param_syms)
    
    # Compute the Jacobian 
    jacobian_matrix = Matrix([model_expr]).jacobian(param_syms)
    jacobian_func = lambdify((x_sym, *param_syms), jacobian_matrix, 'numpy')
    
    # Wrapper to match curve_fit expected format
    def jac_func(x, *params):
        jacobian_values = [jacobian_func(xi, *params) for xi in x]
        return np.array(jacobian_values, dtype=float).squeeze()

    return jac_func

def _normalize_params(params: dict, norm_factor: float,names: list) -> dict:
    """Normalize the parameter values and bounds by the maximum value of the observable."""
    for name in names:
        if params[name]["fix"] is not None:
            params[name]["fix"] = params[name]["fix"]/norm_factor
        else:
            params[name]["guess"] = params[name]["guess"]/norm_factor
            params[name]["bounds"] = (params[name]["bounds"][0]/norm_factor,params[name]["bounds"][1]/norm_factor)
    return params

def _unnormalize_popt(popt,param_order,norm_factor,names):
    for name in names:
        index = param_order.index(name)
        popt[index] = popt[index]*norm_factor
    return popt

def calc_weights_dt(t: np.ndarray):
    """
    Calculate weights as the inverse of the differences in time points

    Parameters
    ----------
    t : np.ndarray
        Array of time points.

    Returns
    -------
    weights : np.ndarray
        Array of weights. May be used as sigma in fitting functions.
    """
    # 
    dt = np.diff(t, prepend=t[1]-t[0])
    weights = 1 / dt
    weights /= np.sum(weights)  # Normalize 
    return weights

def detect_bad_fit(result: ModelResult, atol: float = 1e-10) -> tuple[bool, str]:
    """
    Detects issues with the fit result from lmfit.

    Parameters
    ----------
    result : ModelResult
        The result of the fitting operation from lmfit.

    Returns
    -------
    bad : bool
    message : str
    """
    bad = False
    message = ""

    # Extract fitted data and observed data
    fitted_data = result.best_fit
    y_obs = result.data
    residuals = result.residual

    # Check for constant output
    if np.allclose(fitted_data, fitted_data[0], atol=atol):
        bad = True
        message += f"\n\tModel fit y-values were all within {atol:.2e} of each other."

    # Residual analysis
    if np.any(np.abs(residuals) > 10 * np.std(y_obs)):
        bad = True
        message += "\n\tResiduals are too large."

    # Check if parameters are at bounds
    for name, param in result.params.items():
        if not param.vary:
            continue
        if np.isclose(param.value, param.min):
            bad = True
            message += f"\n\tParameter {name} is at or near its lower bound: {param.min}."
        elif np.isclose(param.value, param.max):
            bad = True
            message += f"\n\tParameter {name} is at or near its upper bound: {param.max}."

    # Check standard errors
    if result.errorbars:
        param_errors = {name: param.stderr for name, param in result.params.items() if param.stderr is not None}
        for name, param in result.params.items():
            if name in param_errors and param_errors[name] > abs(param.value):
                bad = True
                message += f"\n\tLarge standard errors relative to parameter values:"
                message += f"\n\t\t{name}: {param.value:.2e} ± {param_errors[name]:.2e}"

    # Check R^2
    ss_res = np.sum(residuals**2)
    ss_tot = np.sum((y_obs - np.mean(y_obs))**2)
    r_squared = 1 - (ss_res / ss_tot)
    if r_squared < 0.5:
        bad = True
        message += f"\n\tLow R² value: {r_squared:.2e}"

    return bad, message

def calc_nrmse(y_exp: np.ndarray, y_pred: np.ndarray):
    """
    Calculate the Normalized Root Mean Squared Error (NRMSE) between two arrays.

    Parameters
    ----------
    y_exp : np.ndarray
        Array of experimental values.
    y_pred : np.ndarray
        Array of predicted values.

    Returns
    -------
    float
        NRMSE value between 0 and 1, where 1 is a perfect match.
    """
    mse = np.mean((y_exp - y_pred) ** 2)
    rmse = np.sqrt(mse)
    range_y = np.ptp(y_exp)  # Equivalent to max(y_exp) - min(y_exp)
    nrmse = 1 - rmse / range_y
    return nrmse




import colorsys
import numpy as np 

def assign_colors_to_species(schemes: dict, saturation_range: tuple = (0.5, 0.7), lightness_range: tuple = (0.3, 0.4), 
                            method: str = None, offset: float = 0, overwrite_existing=False, seed: int = None):
    """
    Assigns a distinct and aesthetically pleasing color to each species in a dictionary or a single kinetic scheme.
    Uses either a fixed or golden ratio based distribution for hues. Optionally seeds the randomness for consistent results.

    Args:
    schemes (dict): Dictionary of kinetic scheme dictionaries or a single kinetic scheme dictionary.
    saturation_range (tuple): Min and max saturation values.
    lightness_range (tuple): Min and max lightness values.
    method (str): "GR" for golden ratio hue distribution; None for linear distribution.
    offset (float): Offset value for the hues.
    overwrite_existing (bool): If True, overwrite existing colors; if False, assign colors only to species without colors.
    seed (int): Seed for random number generator for reproducible color variations.

    Returns:
    dict: Updated schemes with assigned colors. Edits original input dict.
    """
    #TODO: handle list of schemes and NState class
    if not isinstance(schemes, dict):
        raise ValueError("Input should be a dictionary of schemes or a single scheme formatted as a dictionary.")
    
    single_scheme = False
    if 'species' in next(iter(schemes.keys())):
        single_scheme = True
        schemes = {'single_scheme': schemes}

    # Retrieve unique species into a list. Don't use a set to preserve order.
    unique_species = []
    seen_species = set()
    for scheme in schemes.values():
        for species in scheme["species"].keys():
            if species not in seen_species:
                unique_species.append(species)
                seen_species.add(species)

    golden_ratio_conjugate = 0.618033988749895
    hue = 0

    n_colors = len(unique_species)
    color_mapping = {}
    hues = np.linspace(0, 1, n_colors, endpoint=False)

    # Add existing colors to mapping
    if not overwrite_existing:
        for species in unique_species:
            for scheme in schemes.values():
                if species in scheme["species"] and "color" in scheme["species"][species]:
                    color_mapping[species] = scheme["species"][species]["color"]
                    break

    for i, species in enumerate(unique_species):
        if not overwrite_existing and species in color_mapping:
            continue

        if method == "GR":
            hue += golden_ratio_conjugate + offset
        else:
            hue = hues[i] + offset
        hue %= 1

        np.random.seed(seed) # seed=None will try to read data from /dev/urandom (or the Windows analogue) if available or seed from the clock otherwise
        lightness = np.random.uniform(*lightness_range)
        np.random.seed(seed)
        saturation = np.random.uniform(*saturation_range)

        r, g, b = colorsys.hls_to_rgb(hue, lightness, saturation)
        hex_color = f'#{int(r * 255):02x}{int(g * 255):02x}{int(b * 255):02x}'
        color_mapping[species] = hex_color

    for scheme in schemes.values():
        for species in scheme["species"].keys():
            if overwrite_existing or "color" not in scheme["species"][species]:
                scheme["species"][species]["color"] = color_mapping.get(species, scheme["species"][species].get("color"))

    if single_scheme:
        schemes = schemes['single_scheme']
    return schemes


from datetime import datetime
import sys 

class Logger:
    #TODO: add history with tags of the method that made those, eg system.log.history['simulator']
    #TODO: format log to have method: message
    def __init__(self, quiet=False, logfilename=None):
        self.quiet = quiet
        self.logfilename = logfilename
        self.file_handle = None
        if self.logfilename:
            try:
                self.file_handle = open(self.logfilename, 'a')  # Open the file in append mode
            except IOError as e:
                print(f"Failed to open log file {self.logfilename}: {e}", file=sys.stderr)
                self.file_handle = None
        self._log(f"{datetime.now()} -- Logging started", "INFO")

    def _log(self, log_message, level):
        if not (self.quiet or level == "WARNING") and level != "ERROR":
            print(log_message)  # end='' to avoid double newlines
        if self.file_handle:
            try:
                self.file_handle.write(f"{log_message}\n")
                self.file_handle.flush()
            except IOError as e:
                print(f"Failed to write to log file: {e}", file=sys.stderr)
        if level == "ERROR":
            raise Exception(log_message)

    def info(self, message):
        self._log(message, "INFO")

    def warning(self, message):
        self._log(message, "WARNING")
    
    def error(self, message):
        self._log(message, "ERROR")

    def close(self):
        if self.file_handle:
            try:
                self.file_handle.close()
            except IOError as e:
                print(f"Failed to close log file: {e}", file=sys.stderr)
            finally:
                self.file_handle = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def __del__(self):
        self.close()



import numpy as np
from lmfit import Parameters
from lmfit.model import ModelResult
from typing import Union
from .....utils.helpers import update_dict_with_subset
from .....utils.fitting import general_fit, merge_params

#TODO: fit to scheme. meaning yuo make a scheme without values for the transitions and fit it to occ data to see what values of rates satisfy curve

def occ_final_wrt_t(t,kobs,Etot,uplim=1) -> np.ndarray:
    '''
    Calculate the occupancy of final occupancy (Occ_cov) with respect to time.

    Parameters
    ----------
    t : np.ndarray
        Array of timepoints.
    kobs : float
        Observed rate constant.
    Etot : float
        Total concentration of E across all species.
    uplim : float, optional
        Upper limit scalar of the curve. The fraction of total E typically. Default is 1, i.e., 100%.

    Returns
    -------
    np.ndarray
        Occupancy of final occupancy (Occ_cov).
    '''
    return uplim * Etot * (1 - np.e**(-kobs * t))

def kobs_uplim_fit_to_occ_final_wrt_t(t, occ_final, nondefault_params: Union[dict,Parameters] = None, xlim: tuple = None, 
                                        weights_kde=False, weights: np.ndarray = None, verbosity=2, **kwargs) -> ModelResult:
    '''
    Fit kobs to the first order occupancy over time.

    Parameters
    ----------
    t : np.ndarray
        Array of timepoints.
    occ_final : np.ndarray
        Array of observed occupancy, i.e. concentration.
    nondefault_params : dict or Parameters, optional
        A structured dictionary of parameters with 'value','vary', and 'bound' keys or a lmfit.Parameters object.
        Defaults:
        ```python
        # Observed rate constant
        default_params.add('kobs', value=0.01, vary=True, min=0, max=np.inf)
        # Total concentration of E over all species
        default_params.add('Etot', value=1, vary=False, min=0, max=np.inf)
        # Scales the upper limit of the curve
        default_params.add('uplim', value=1, vary=True, min=0, max=np.inf)
        ```
        Example dict of nondefaults:
        ```python
        nondefault_params = {
            "Etot": {"vary": False, "value": 0.5},  
            "uplim": {"vary": False},    
        }
        ```
    xlim : tuple, optional
        Limits for the time points considered in the fit (min_t, max_t).
    weights_kde : bool, optional
        If True, calculate the density of the x-values and use the normalized reciprocol as weights. Similar to 1/sigma for scipy.curve_fit.
        Helps distribute weight over unevenly-spaced points. Default is False.
    weights : np.ndarray, optional
        weights parameter for fitting. This argument is overridden if weights_kde=True. Default is None.
    verbosity : int, optional
        0: print nothing. 1: print upon bad fit. 2: print always. Default is 2.
    kwargs : dict, optional
        Additional keyword arguments to pass to the lmfit Model.fit function.

    Returns
    -------
    lmfit.ModelResult
        The result of the fitting operation from lmfit.

    '''

    default_params = Parameters()
    default_params.add('kobs', value=0.01, vary=True, min=0, max=np.inf)
    default_params.add('Etot', value=1, vary=False, min=0, max=np.inf)
    default_params.add('uplim', value=1, vary=True, min=0, max=np.inf)

    lm_params = merge_params(default_params, nondefault_params)

    return general_fit(occ_final_wrt_t, t, occ_final, lm_params, xlim=xlim, weights_kde=weights_kde, weights=weights, verbosity=verbosity, **kwargs)




def occ_total_wrt_t(t,kobs,concI0,KI,Etot,uplim=1):
    '''
    Calculates pseudo-first-order total occupancy of all bound states, 
    assuming fast reversible binding equilibrated at t=0.

    Parameters
    ----------
    t : np.ndarray
        Array of timepoints.
    kobs : float
        Observed rate constant.
    concI0 : float
        Initial concentration of the (saturating) inhibitor.
    KI : float
        Inhibition constant, where kobs = kinact/2, analogous to K_M, K_D, and K_A. 
        Must be in the same units as concI0.
    Etot : float
        Total concentration of E across all species.
    uplim : float, optional
        Upper limit scalar of the curve. The fraction of total E typically. Default is 1, i.e., 100%.

    Returns
    -------
    np.ndarray
        Occupancy of total occupancy (Occ_tot).
    '''

    FO = 1 / (1 + (KI / concI0)) # Equilibrium occupancy of reversible portion
    return uplim * Etot * (1 - (1 - FO) * np.exp(-kobs * t))

def kobs_KI_uplim_fit_to_occ_total_wrt_t(t: np.ndarray, occ_tot: np.ndarray, nondefault_params: Union[dict,Parameters] = None, xlim: tuple = None, 
                                        weights_kde=False, weights: np.ndarray = None, verbosity=2, **kwargs) -> ModelResult:
    '''
    Fit kobs and KI to the total occupancy of all bound states over time, 
    assuming fast reversible binding equilibrated at t=0.

    Parameters
    ----------
    t : np.ndarray
        Array of timepoints.
    occ_tot : np.ndarray
        Array of total bound enzyme population. All occupied states.
    nondefault_params : dict or Parameters, optional
        A structured dictionary of parameters with 'value','vary', and 'bound' keys or a lmfit.Parameters object.
        Defaults:
        ```python
        # Observed rate constant
        default_params.add('kobs', value=0.01, vary=True, min=0, max=np.inf)
        # Initial concentration of the (saturating) inhibitor
        default_params.add('concI0', value=100, vary=True, min=0, max=np.inf)
        # Inhibition constant where kobs = kinact/2.
        default_params.add('KI', value=10, vary=True, min=0, max=np.inf)
        # Total concentration of E across all species
        default_params.add('Etot', value=1, vary=False, min=0, max=np.inf)
        # Scales the upper limit of the curve
        default_params.add('uplim', value=1, vary=True, min=0, max=np.inf)      
        ```
        Example dict of nondefaults:
        ```python
        nondefault_params = {
            "Etot": {"vary": False, "value": 0.5},  
            "uplim": {"vary": False},    
        }
        ```
    xlim : tuple, optional
        Limits for the time points considered in the fit (min_t, max_t).
    weights_kde : bool, optional
        If True, calculate the density of the x-values and use the normalized reciprocol as weights. Similar to 1/sigma for scipy.curve_fit.
        Helps distribute weight over unevenly-spaced points. Default is False.
    weights : np.ndarray, optional
        weights parameter for fitting. This argument is overridden if weights_kde=True. Default is None.
    verbosity : int, optional
        0: print nothing. 1: print upon bad fit. 2: print always. Default is 2.
    kwargs : dict, optional
        Additional keyword arguments to pass to the lmfit Model.fit function.

    Returns
    -------
    lmfit.ModelResult
        The result of the fitting operation from lmfit.

    '''
    default_params = Parameters()
    default_params.add('kobs', value=0.01, vary=True, min=0, max=np.inf)
    default_params.add('concI0', value=100, vary=True, min=0, max=np.inf)
    default_params.add('KI', value=10, vary=True, min=0, max=np.inf)
    default_params.add('Etot', value=1, vary=False, min=0, max=np.inf)
    default_params.add('uplim', value=1, vary=True, min=0, max=np.inf)

    lm_params = merge_params(default_params, nondefault_params)
    return general_fit(occ_total_wrt_t, t, occ_tot, lm_params, xlim=xlim, weights_kde=weights_kde, weights=weights, verbosity=verbosity, **kwargs)

def kobs_wrt_concI0(concI0,KI,kinact,n=1): 
    '''
    Calculates the observed rate constant kobs with respect to the initial 
    concentration of the inhibitor using a Michaelis-Menten-like equation.

    Parameters
    ----------
    concI0 : np.ndarray
        Array of initial concentrations of the inhibitor.
    KI : float
        Inhibition constant, analogous to K_M, K_D, and K_A, where kobs = kinact/2.
    kinact : float
        Maximum potential rate of covalent bond formation.
    n : float, optional
        Hill coefficient, default is 1.

    Returns
    -------
    np.ndarray
        Array of kobs values, the first order observed rate constants of inactivation, 
        with units of inverse time.
    
    Notes
    -----
    Assumes that concI is constant over the timecourses where kobs is calculated. 
    '''
    return kinact / (1 + (KI / concI0)**n)

def KI_kinact_n_fit_to_kobs_wrt_concI0(concI0: np.ndarray, kobs: np.ndarray, nondefault_params: Union[dict,Parameters] = None, xlim: tuple = None, 
                                        weights_kde=False, weights: np.ndarray = None, verbosity=2, **kwargs) -> ModelResult:
    """
    Fit parameters (KI, kinact, n) to kobs with respect to concI0 using 
    a structured dictionary for parameters.

    Parameters
    ----------
    concI0 : np.ndarray
        Array of initial concentrations of the inhibitor.
    kobs : np.ndarray
        Array of observed rate constants.
    nondefault_params : dict or Parameters, optional
        A structured dictionary of parameters with 'value','vary', and 'bound' keys or a lmfit.Parameters object.
        Defaults:
        ```python
        default_params.add('KI', value=100, vary=True, min=0, max=np.inf)
        default_params.add('kinact', value=0.01, vary=True, min=0, max=np.inf)
        default_params.add('n', value=1, vary=False, min=0, max=np.inf)   
        ```
        Example dict of nondefaults:
        ```python
        nondefault_params = {
            "n": {"vary": True},    
        }
        ```
    xlim : tuple, optional
        Limits for the time points considered in the fit (min_t, max_t).
    weights_kde : bool, optional
        If True, calculate the density of the x-values and use the normalized reciprocol as weights. Similar to 1/sigma for scipy.curve_fit.
        Helps distribute weight over unevenly-spaced points. Default is False.
    weights : np.ndarray, optional
        weights parameter for fitting. This argument is overridden if weights_kde=True. Default is None.
    verbosity : int, optional
        0: print nothing. 1: print upon bad fit. 2: print always. Default is 2.
    kwargs : dict, optional
        Additional keyword arguments to pass to the lmfit Model.fit function.

    Returns
    -------
    lmfit.ModelResult
        The result of the fitting operation from lmfit.
    
    Assumes that concI is constant over the timecourses where kobs is calculated. 
    """
    default_params = Parameters()
    default_params.add('KI', value=100, vary=True, min=0, max=np.inf)
    default_params.add('kinact', value=0.01, vary=True, min=0, max=np.inf)
    default_params.add('n', value=1, vary=True, min=0, max=np.inf)

    lm_params = merge_params(default_params, nondefault_params)
    return general_fit(kobs_wrt_concI0, concI0, kobs, lm_params, xlim=xlim, weights_kde=weights_kde, weights=weights, verbosity=verbosity, **kwargs)

def dose_response(dose: np.ndarray,Khalf: float, kinact: float, t: float, n=1): 
    '''
    Calculates the Hill equation for a dose-response curve.

    Parameters
    ----------
    dose : np.ndarray
        Array of input concentrations of the ligand.
    Khalf : float
        The dose required for half output response.
    t : float
        The endpoint used of dosing. 
    kinact : 
        The apparent maximal rate constant of inactivation. 
    n : float, optional
        Hill coefficient, default is 1.

    Returns
    -------
    np.ndarray
        The fraction of the responding population.
    '''
    return (1 - np.exp(-kinact * t)) / (1 + (Khalf / dose)**n)

def dose_response_fit(dose: np.ndarray, response: np.ndarray, nondefault_params: Union[dict,Parameters] = None, xlim: tuple = None, 
                                        weights_kde=False, weights: np.ndarray = None, verbosity=2, **kwargs) -> ModelResult:
    """
    Fit parameters (Khalf, kinact, n) to response with respect to dose using 
    a structured dictionary for parameters.

    Parameters
    ----------
    dose : np.ndarray
        Array of input concentrations of the ligand.
    response : np.ndarray
        Array of the fraction of the responding population.
    nondefault_params : dict or Parameters, optional
        A structured dictionary of parameters with 'value','vary', and 'bound' keys or a lmfit.Parameters object.
        Defaults:
        ```python
        default_params.add('Khalf', value=100, vary=True, min=0, max=np.inf)
        default_params.add('kinact', value=0.01, vary=True, min=0, max=np.inf)
        default_params.add('t', value=3600, vary=False)
        default_params.add('n', value=1, vary=True, min=0, max=np.inf)
        ```
        Example dict of nondefaults:
        ```python
        nondefault_params = {
            "n": {"vary": False},    
        }
        ```
    xlim : tuple, optional
        Limits for the time points considered in the fit (min_t, max_t).
    weights_kde : bool, optional
        If True, calculate the density of the x-values and use the normalized reciprocol as weights. Similar to 1/sigma for scipy.curve_fit.
        Helps distribute weight over unevenly-spaced points. Default is False.
    weights : np.ndarray, optional
        weights parameter for fitting. This argument is overridden if weights_kde=True. Default is None.
    verbosity : int, optional
        0: print nothing. 1: print upon bad fit. 2: print always. Default is 2.
    kwargs : dict, optional
        Additional keyword arguments to pass to the lmfit Model.fit function.

    Returns
    -------
    lmfit.ModelResult
        The result of the fitting operation from lmfit.
    """
    default_params = Parameters()
    default_params.add('Khalf', value=100, vary=True, min=0, max=np.inf)
    default_params.add('kinact', value=0.01, vary=True, min=0, max=np.inf)
    default_params.add('t', value=3600, vary=False)
    default_params.add('n', value=1, vary=True, min=0, max=np.inf)

    lm_params = merge_params(default_params, nondefault_params)
    return general_fit(dose_response, dose, response, lm_params, xlim=xlim, weights_kde=weights_kde, weights=weights, verbosity=verbosity, **kwargs)

class Params:
    """
    Common place for parameters found in covalent inhibition literature.
    """
    @staticmethod
    def Ki(kon, koff):
        """
        Ki (i.e. inhib. dissociation constant, Kd) calculation

        Parameters
        ----------
        kon : float
            On-rate constant (CONC^-1*TIME^-1).
        koff : float
            Off-rate constant (TIME^-1).

        Returns
        -------
        float
            The calculated inhib. dissociation constant (Ki).

        Notes
        -----
        The inhib. dissociation constant (Ki) is calculated as koff / kon.
        """
        return koff / kon

    @staticmethod
    def KI(kon, koff, kinact):
        """
        KI (i.e. inhibition constant, KM, KA, Khalf, KD (not to be confused with Kd)) calculation.
        Numerically, this should be the concentration of inhibitor that yields kobs = 1/2*kinact.

        Parameters
        ----------
        kon : float
            On-rate constant (CONC^-1*TIME^-1).
        koff : float
            Off-rate constant (TIME^-1).
        kinact : float
            Inactivation (last irreversible step) rate constant.

        Returns
        -------
        float
            The calculated inhibition constant (KI).

        Notes
        -----
        The inhibition constant (KI) is calculated as (koff + kinact) / kon.
        """
        return (koff + kinact) / kon

    @staticmethod
    def kinact_app(t: float, Prob_cov: float):
        """
        The apparent maximal rate constant of inactivation, calculated from a single timepoint.

        Parameters
        ----------
        t : float
            Endpoint for the dosing.
        Prob_cov : float
            Probability (ie fraction) of the covalently-bound state.

        Notes
        -----
        The data must be from a timepoint where kobs ~= kinact due to large [I].

        Can be obtained from the upper limit of a fitted dose response curve. 
        """
        return -np.log(1-Prob_cov)/t
    
    @staticmethod
    def CE(kon, koff, kinact):
        """
        Covalent efficiency (i.e. specificity, potency) calculation (kinact/KI).

        Parameters
        ----------
        kon : float
            On-rate constant (CONC^-1*TIME^-1).
        koff : float
            Off-rate constant (TIME^-1).
        kinact : float
            Inactivation (last irreversible step) rate constant.

        Returns
        -------
        float
            The calculated covalent efficiency (kinact/KI).

        Notes
        -----
        The covalent efficiency is calculated as the ratio of 
        the inactivation rate constant (kinact) to the inhibition constant (KI).
        """
        return kinact/Parameters.KI(kon, koff, kinact)

class Experiments:
    """
    Common place for experimental setups in covalent inhibition literature.
    """
    #TODO: timecourse and KI/kinact 
    @staticmethod
    def timecourse(t,system):
        return NotImplementedError()


from setuptools import setup, find_packages

with open('requirements.txt') as f:
    required = f.read().splitlines()

setup(
    name='GeKiM', 
    version='0.5.1', 
    author='Kyle Ghaby', 
    author_email='kyleghaby@gmail.com',  
    description='Generalized Kinetic Modeler: A Python package for modeling arbitrary kinetic schemes.',  
    long_description=open('README.md').read(),  
    long_description_content_type='text/markdown',
    url='https://github.com/kghaby/GeKiM', 
    packages=find_packages(), 
    include_package_data=True,
    install_requires=required,
    classifiers=[
        'Development Status :: 4 - Beta', 
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Chemistry',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Programming Language :: Python :: 3.9',
    ],
    python_requires='>=3.9', 
)



class Params:
    @staticmethod
    def KM(kon, koff, kcat):
        """
        KM (i.e. Michaelis-Menten constant, KA, Khalf, KD (not to be confused with Kd)) calculation.
        
        Parameters
        ----------
        kon : float
            On-rate constant (CONC^-1*TIME^-1)
        koff : float
            Off-rate constant (TIME^-1)
        kcat : float
            Irreversible catalysis rate constant
        
        Returns
        -------
        float
            The calculated Michaelis-Menten constant (KM)
        """
        return (koff + kcat) / kon


from abc import ABC, abstractmethod
import numpy as np
from itertools import product
import weakref
from ..schemes.n_state import NState

class BaseSimulator(ABC):
    def __init__(self, system: NState):
        """
        Initialize the simulator.

        Parameters
        ----------
        system : NState
            The system object. 

        Notes
        -----
        The input NState instance, `system`, will be modified directly 
            by the simulator, whether the simulator is added as an attribute to `system` or not.
            
        Initializes the `simin` and `simout` dictionaries of system, species, and transitions objects.
        """
        system.log.info(f"Setting up solver {self.__class__}...\n")
        self.system = weakref.proxy(system)

        self.system.simin = {}
        self.system.simout = {}
        for _, sp_data in self.system.species.items():
            sp_data.simin = {}
            sp_data.simout = {}
        for _, tr_data in self.system.transitions.items():
            tr_data.simin = {}
            tr_data.simout = {}

        self.setup()

    @abstractmethod
    def setup(self):
        """
        Perform any necessary setup specific to the simulator.
        This method should be overridden by subclasses.
        """
        pass

    @abstractmethod
    def simulate(self, output_raw=False, **kwargs):
        """
        Run the simulation.
        This method should be overridden by subclasses.
        """
        pass

    def _make_y0_mat(self):
        """
        Create a matrix of initial conditions for the system.
        """
        combinations = product(*(
            np.atleast_1d(np.atleast_1d(sp_data.y0).flatten())
            for _, sp_data in self.system.species.items()
        ))
        y0_mat = np.vstack([comb for comb in combinations])
        return y0_mat






import numpy as np
from scipy.integrate import solve_ivp
from sympy import symbols, Matrix, prod, pretty, zeros, lambdify
from .base_simulator import BaseSimulator

class ODESolver(BaseSimulator):
    """
    Parameters
    ----------
    system : NState
        The system object.

    Notes
    -----
    The input NState instance, `system`, will be modified directly by the simulator, 
    whether the simulator is added as an attribute to `system` or not. 

    Use as a simulator for NState class. Will generate and solve systems deterministically.

    During initialization, the following keys (and values) are added to the 
    `self.system.simin` dict:

        unit_sp_mat : np.ndarray
            Unit species matrix.
        stoich_mat : np.ndarray
            Stoichiometry matrix.
        stoich_reactant_mat : np.ndarray
            Stoichiometry reactant matrix.
        k_vec : np.ndarray
            Rate constant vector.
        k_diag : np.ndarray
            Diagonal matrix of rate constants.
        rates_sym : sympy.Matrix
            Species rate vectors with symbolic species and rate constants.
        rates_numk : sympy.Matrix
            Species rate vector with symbolic species and numerical rate constants.
        J_sym : sympy.Matrix
            Jacobian matrix with symbolic species and rate constants.
        J_symsp_numtr : sympy.Matrix
            Jacobian matrix with symbolic species and numerical rate constants.
        J_func_wrap : Callable
            Wrapped numerical Jacobian function that accepts t, y as arguments.
    """

    def setup(self):
        """
        (Re)initialize the simulator by generating the necessary matrices and symbolic expressions, 
        and saving them to the system and species simin dictionaries.
        """
        
        self._generate_matrices_for_rates_func() 

        # Rate laws 
        self._generate_rates_sym() 
        self.log_rates()
        #self.lambdify_sym_rates() # Overwrites self._rates_func with a lambdified version of self.system.simin["rates_sym"]

        # Jacobian
        self._generate_jac() 
        self.log_jac()

    def _generate_matrices_for_rates_func(self):
        """
        Notes
        -----
        Generates 
        - unit species matrix (`self.system.simin["unit_sp_mat"]`), 
        - stoichiometry matrix (`self.system.simin["stoich_mat"]`), 
        - stoichiometry reactant matrix (`self.system.simin["stoich_reactant_mat"]`), and 
        - rate constant vector (`self.system.simin["k_vec"]`) 
            and diagonal matrix (`self.system.simin["k_diag"]`).

        Rows are transitions, columns are species.

        Used for solving rates.
        """

        n_species = len(self.system.species)
        n_transitions = len(self.system.transitions)
        self.system.simin["unit_sp_mat"] = np.eye(n_species, dtype=int)

        # Initialize matrices
        self.system.simin["stoich_reactant_mat"] = np.zeros((n_transitions, n_species))
        self.system.simin["stoich_mat"] = np.zeros((n_transitions, n_species))
        self.system.simin["k_vec"] = np.zeros(n_transitions)

        # Fill stoich matrices and k vector
        for tr_name, tr in self.system.transitions.items():
            tr_idx = tr.index
            self.system.simin["k_vec"][tr_idx] = tr.k
            reactant_vec = np.sum([self.system.simin["unit_sp_mat"][self.system.species[name].index] * coeff for name, coeff in tr.source],axis=0)
            product_vec = np.sum([self.system.simin["unit_sp_mat"][self.system.species[name].index] * coeff for name, coeff in tr.target],axis=0)
            
            self.system.simin["stoich_reactant_mat"][tr_idx, :] = reactant_vec  
            #self._stoich_product_mat[tr_idx, :] = product_vec   # not used
            self.system.simin["stoich_mat"][tr_idx] = product_vec - reactant_vec

        self.system.simin["k_diag"] = np.diag(self.system.simin["k_vec"])

        return
    
    def _rates_func(self, t, conc):
        """
        Calculate the rate vector at time t and concentration vector conc.

        Notes
        -----
        Cannot model rates that are not simple power laws (eg dynamic inhibition, cooperativity, time dependent params). 
        But most of these can be baked in on the schematic level I think. 
        """
        #TODO: Use higher dimensionality conc arrays to process multiple input concs at once? 
        C_Nr = np.prod(np.power(conc, self.system.simin["stoich_reactant_mat"]), axis=1) # state dependencies
        N_K = np.dot(self.system.simin["k_diag"],self.system.simin["stoich_mat"]) # interactions
        rates = np.dot(C_Nr,N_K)
        return rates
    
    def _generate_rates_sym(self):
        """
        Notes
        -----
        Generates 
        - Species rate vectors with symbolic species and rate constants 
        - Species rate vectors with symbolic species and numerical rate constants

        Adds these key/value pairs to both system and species simin dictionaries.
        """
        # Generate rate's with symbolic species and rate constants 
        rates_sym = Matrix([0] * len(self.system.species))
        for tr_name, tr in self.system.transitions.items():
            unscaled_rate = self.system.transitions[tr_name].sym * prod(self.system.species[sp_name].sym**coeff for sp_name, coeff in tr.source)
            for sp_name, coeff in tr.source:
                rates_sym[self.system.species[sp_name].index] -= coeff * unscaled_rate
            for sp_name, coeff in tr.target:
                rates_sym[self.system.species[sp_name].index] += coeff * unscaled_rate
        self.system.simin["rates_sym"] = rates_sym 
        self.system.log.info("Saved rates_sym and rates_numk into system simin dict (SYSTEM.simin).")

        # Substitute rate constant symbols for values 
        tr_sym2num = {symbols(name): tr.k for name, tr in self.system.transitions.items()}
        self.system.simin["rates_numk"] = self.system.simin["rates_sym"].subs(tr_sym2num)

        # Assign each rate law to the respective species
        for sp_name, sp_data in self.system.species.items():
            sp_data.simin["rate_sym"] = self.system.simin["rates_sym"][sp_data.index]
            sp_data.simin["rate_numk"] = self.system.simin["rates_numk"][sp_data.index]
        self.system.log.info("Saved rate_sym and rate_numk into species simin dict (SYSTEM.species[NAME].simin).")

        return 
    
    def lambdify_sym_rates(self):
        """
        Convert the symbolic rate vector (with numerical rate constants) into a numerical function.
    
        Notes
        -----
        This overwrites the unsymbolic self._rates_func function. It is slightly slower but more flexible.

        Can be used to simulate with custom symbolic rate funcs for each species instead of all following the same ODE paradigm.
        """
        #TODO: test custom funcs and refactor as needed to make sure its easier and more robust to use (ie make sure the appropriate setup steps are reran
        # currently uses the system.simin, not species specific. would need to remake system.simin, or alter rate func at the generate_rates_sym step
        species_vec = Matrix([self.system.species[sp_name].sym for sp_name in self.system.species])
        rates_func = lambdify(species_vec, self.system.simin["rates_numk"], 'numpy')
        self._rates_func = lambda t, y: rates_func(*y).flatten()

        # Re-generate the Jacobian matrix in case this function is called after the Jacobian has been generated, like for custom functions
        #self.generate_jac()
        return

    def log_rates(self,force_print=False):
        """
        Log the symbolic rates using the system's logger.

        Parameters
        ----------
        force_print : bool, optional
            If True, the rates will be printed to the console. Default is False.
        """
        rate_dict = {}
        max_header_length = 0

        # Find the max length for the headers
        for sp_name in self.system.species:
            header_length = len(f"d[{sp_name}]/dt")
            max_header_length = max(max_header_length, header_length)

        # Write eqn headers and rate laws
        for sp_name in self.system.species:
            rate_dict[sp_name] = [f"d[{sp_name}]/dt".ljust(max_header_length) + " ="]

        for tr_name, tr in self.system.transitions.items():
            # Write rate law
            rate = f"{tr_name} * " + " * ".join([f"{sp_name}^{coeff}" if coeff != 1 else f"{sp_name}" for sp_name,coeff in tr.source])
            rate = rate.rstrip(" *")  # Remove trailing " *"

            # Add rate law to the eqns
            for sp_name,coeff in tr.source:
                term = f"{coeff} * {rate}" if coeff != 1 else rate
                rate_dict[sp_name].append(f" - {term}")

            for sp_name,coeff in tr.target:
                term = f"{coeff} * {rate}" if coeff != 1 else rate
                rate_dict[sp_name].append(f" + {term}")

        # Construct the final string
        rate_log = "\nRates:\n\n"
        for sp_name, eqn_parts in rate_dict.items():
            # Aligning '+' and '-' symbols
            eqn_header = eqn_parts[0]
            terms = eqn_parts[1:]
            aligned_terms = [eqn_header + " " + terms[0]] if terms else [eqn_header]
            aligned_terms += [f"{'':>{max_header_length + 3}}{term}" for term in terms[1:]]
            formatted_eqn = "\n".join(aligned_terms)
            rate_log += formatted_eqn + '\n\n'

        self.system.log.info(rate_log)
        if force_print:
            print(rate_log)
        return

    def _generate_jac(self):
        """
        Notes
        -----
        Generates
         - the Jacobian matrix with symbolic species and rate constants (self.system.simin["J_sym"])
         - the Jacobian matrix with symbolic species and numerical rate constants (self.system.simin["J_symsp_numtr"])
         - a wrapped numerical Jacobian function (self.system.simin["J_func_wrap"]) that accepts t,y as arguments

        The wrapped numerical function is time-independent even though it accepts t as an argument!

        The Jacobian matrix here represents the first-order partial derivatives of the rate of 
            change equations for each species with respect to all other species in the system.
        """
        
        species_vec = Matrix([self.system.species[sp_name].sym for sp_name in self.system.species])

        # Symbolic Jacobian
        self.system.simin["J_sym"] = self.system.simin["rates_sym"].jacobian(species_vec)

        # Numerical Jacobian
        self.system.simin["J_symsp_numtr"] = self.system.simin["rates_numk"].jacobian(species_vec) # Symbolic species, numeric transition rate constants
        J_func = lambdify(species_vec, self.system.simin["J_symsp_numtr"], 'numpy') # Make numerical function
        self.system.simin["J_func_wrap"] = lambda t, y: J_func(*y) # Wrap J_func so that t,y is passed to the function to be compatible with solve_ivp
        return
      
    def log_jac(self,force_print=False):
        """
        Log the symbolic representation of the Jacobian matrix using the system's logger.

        Notes
        -----
        The logged Jacobian includes row and column labels, but `SYSTEM.simin["J_sym"]` does not.

        The Jacobian matrix here represents the first-order partial derivatives of the rate of 
            change equations for each species with respect to all other species in the system.
        """
        n_species = len(self.system.species)
        J_log = zeros(n_species + 1, n_species + 1)
        J_log[1:, 1:] = self.system.simin["J_sym"]

        for sp_name,sp_data in self.system.species.items():
            J_log[0, sp_data.index+1] = symbols(sp_name)
            J_log[sp_data.index+1, 0] = symbols(f'd[{sp_name}]/dt')
        J_log[0,0] = symbols("_")

        J_log_str = "Jacobian (including row and column labels):\n"
        J_log_str += pretty((J_log),use_unicode=True)
        J_log_str += "\n"
        self.system.log.info(J_log_str)
        if force_print:
            print(J_log_str)
        return

    def simulate(self, t_eval: np.ndarray = None, t_span: tuple = None, method='BDF', rtol=1e-3, atol=0, dense_output=False, output_raw=False, **kwargs):
        """
        Solve the differential equations of species concentration wrt time for the system. 

        Parameters
        ----------
        output_raw : bool, optional
            If True, return raw solver output.
        t_eval : np.ndarray, optional
            Time points for rate solutions.
        t_span : tuple, optional
            Time span for rate solutions.
        method : str, optional
            Integration method, default is 'BDF'.
        rtol : float, optional
            Relative tolerance for the solver. Default is 1e-3.
        atol : float, optional
            Absolute tolerance for the solver. Default is 0. If atol == 0, atol = 1e-6 * <the smallest nonzero value of y0>.
        dense_output : bool, optional
            If True, save a `scipy.integrate._ivp.common.OdeSolution` instance to `SYSTEM.simout.soln_continuous(t)`.
            If using multiple y0's, this will be a list of instances that share indexing with the other outputs,
            and can be called like `SYSTEM.simout['soln_continuous'][idx](t)`.
            Access a specific species conc like `SYSTEM.simout['soln_continuous'](t)[SYSTEM.species[NAME].index]`.
        **kwargs : dict, optional
            Additional keyword arguments to pass to the solver.

        Notes
        -----
        Saves the following keys to `SYSTEM.simout`:
        - t : np.ndarray
            Time points
        - soln_continuous : scipy.integrate._ivp.common.OdeSolution or list of scipy.integrate._ivp.common.OdeSolution
            Continuous solution function (if `dense_output=True`)

        Saves the following keys to each species' `simout` dictionary attribute:
        - y : np.ndarray
            Concentration vectors

        Returns
        -------
        None or scipy.integrate._ivp.ivp.OdeResult or list of scipy.integrate._ivp.ivp.OdeResult
        """
        #TODO: check how combinations are arranged and make sure its intuitive to separate and use them (ie the indexing is clear)
        #TODO: More analytically approximate the time scale.
            # incorporate the network of transitions, nonlinearity, etc?
            # Linear scaling of the inverse min eigenvalue underestimates when y0E ~= y0I
            # Linearize system then use normal mode frequency of linear system (1/(sqrt(smallest eigenvalue))?
            # needs to be an n-dimensional function, where n is the degree of (non)linearity

        y0_mat = self._make_y0_mat()
        y0_mat_len = len(y0_mat)
        self.system.log.info(f"Solving the timecourse from {y0_mat_len} initial concentration vectors...")

        solns = []
        for y0 in y0_mat:
            if atol == 0:
                atol = np.min(y0[y0 != 0])*1e-6 # smallest nonzero value of the Jacobian
                self.system.log.info(f"\tSetting atol to {atol:.2e}, 1e-6 * <the smallest nonzero value of y0>.")
                
            if t_span is None:
                if t_eval is None:
                    J0 = self.system.simin["J_func_wrap"](None, y0)
                    t_span = self.estimate_t_span(J0)
                    self.system.log.info(f"\tEstimated time scale: {t_span[1]:.2e} (1/<rate constant units>)")
                else:
                    t_span = (t_eval[0], t_eval[-1])

            soln = solve_ivp(self._rates_func, t_span=t_span, y0=y0, method=method, t_eval=t_eval, 
                                rtol=rtol, atol=atol, jac=self.system.simin["J_func_wrap"], dense_output=dense_output, **kwargs) 
                # vectorized=True makes legacy rate func slower bc low len(y0) I think
            if not soln.success:
                raise RuntimeError("FAILED: " + soln.message)
            solns.append(soln)
            
        self.system.log.info("ODE's solved successfully. Saving data...")

        self._process_simouts(solns,y0_mat_len,dense_output=dense_output)

        if output_raw:
            if y0_mat_len == 1:
                self.system.log.info("Returning raw solver output.\n")
                return solns[0]
            self.system.log.info("Returning list of raw solver outputs.\n")        
            return solns
        else:
            self.system.log.info("Not returning raw solver output. Use output_raw=True to return raw data.\n")
            return

    def estimate_t_span(self, J0: np.ndarray) -> tuple[float, float]:
        """
        Estimate the timespan needed for convergence based on the 
        smallest magnitude of the Jacobian eigenvalues at initial conditions
        
        Parameters:
        ----------
        J0 : np.ndarray
            The Jacobian of the initial conditions of the system
        
        Returns:
        -------
        tuple[float, float]
            A tuple containing the estimated timespan needed for convergence.
            The first element of the tuple is the start time (0) and the second
            element is the estimated end time based on the smallest magnitude
            of the Jacobian eigenvalues at the initial conditions.
        
        Raises:
        ------
        ValueError
            If no eigenvalues above the threshold are found, indicating that
            the time scale cannot be estimated.
        """
        eigenvalues = np.linalg.eigvals(J0)
        eigenvalue_threshold = 1e-6 # below 1e-6 is considered insignificant. float32 cutoff maybe
        filtered_eigenvalues = eigenvalues[np.abs(eigenvalues) > eigenvalue_threshold] 
        if filtered_eigenvalues.size == 0:
            raise ValueError("No eigenvalues above the threshold, unable to estimate time scale.")
        naive_time_scale = 1 / (np.abs(filtered_eigenvalues).min())
        est_time_scale = naive_time_scale * 6.5
        est_t_span = (0, est_time_scale) # Start at 0 or np.abs(filtered_eigenvalues).min()?
        return est_t_span
    
    def _process_simouts(self, raw_simouts: list, y0_mat_len: int, dense_output=False):
        if y0_mat_len == 1:
            self.system.simout["t"] = raw_simouts[0].t
            self.system.log.info(f"\tTime saved to SYSTEM.simout['t'] (np.array)")
            self.system.simout["y"] = raw_simouts[0].y
            for _, data in self.system.species.items():
                data.simout["y"] = self.system.simout["y"][data.index]
            self.system.log.info(f"\tConcentrations saved respectively to SYSTEM.species[NAME].simout['y'] (np.array)")
            if dense_output:
                self.system.simout["soln_continuous"] = raw_simouts[0].sol
                self.system.log.info(f"\tSaving continuous solution function to SYSTEM.simout['soln_continuous'](t) (scipy.integrate.ODESolution)")
            else:
                self.system.simout["soln_continuous"] = None
                self.system.log.info("\tNot saving continuous solution. Use dense_output=True to save it to SYSTEM.simout['soln_continuous']")
        else:
            self.system.simout["t"] = [raw_simout.t for raw_simout in raw_simouts] 
            self.system.log.info(f"\t{y0_mat_len} time vectors saved to SYSTEM.simout['t'] (list of np.arrays)")
            self.system.simout["y"] = [raw_simout.y for raw_simout in raw_simouts]
            for _, data in self.system.species.items():
                data.simout["y"] = [y[data.index] for y in self.system.simout["y"]]
            self.system.log.info(f"\t{y0_mat_len} concentration vectors saved respectively to SYSTEM.species[NAME].simout['y'] (list of np.arrays)")
            if dense_output:
                self.system.simout["soln_continuous"] = [raw_simout.sol for raw_simout in raw_simouts] 
                self.system.log.info(f"\tSaving list of continuous solution functions to SYSTEM.simout['soln_continuous'] (list of scipy.integrate.ODESolution's)")
            else:
                self.system.simout["soln_continuous"] = None
                self.system.log.info("\tNot saving continuous solutions. Use dense_output=True to save them to SYSTEM.simout['soln_continuous']")
        return
